##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" Discrete pseudo motor controller configuration related macros"""

__all__ = ["def_discr_pos", "udef_discr_pos", "prdef_discr"]

__docformat__ = 'restructuredtext'

import math
from taurus.core.util.codecs import CodecFactory
from taurus.console.table import Table
from sardana.macroserver.macro import Macro, Type


class DiscretePseudoMotorConfiguration(dict):

    def __init__(self, pseudo_obj, macro):
        self.pseudo = pseudo_obj
        self.macro = macro
        _physical_motor_name = self.pseudo.physical_elements[0]
        self.motor = macro.getMoveable(_physical_motor_name)
        cf = CodecFactory()
        self.json = cf.getCodec('json')
        conf = self.get_configuration()
        self.update(conf)

    def get_configuration(self):
        value = self.pseudo.getAttribute('configuration').read().value
        fmt, data = self.json.decode(('json', value))
        return data

    def has_calibration(self):
        return all(['set' in list(self[x].keys()) for x in list(self.keys())])

    def add_point(self, label, pos, setpos, dmin, dmax):
        point = dict()
        point['pos'] = int(pos)
        label = label.upper()
        # Calculate point calibration if required
        if self.has_calibration():
            # Set to current physical position if no value supplied as argument
            if math.isinf(setpos):
                point['set'] = self.motor.position
            else:
                point['set'] = float(setpos)
            # If point exists, we use current min, max values
            if (label in list(self.keys())
                    and math.isinf(dmin)
                    and math.isinf(dmax)):
                p = self[label]
                min_pos = point['set'] + p['set'] - p['min']
                max_pos = point['set'] + p['set'] - p['max']
            # else, new point has new calibration,
            elif math.isinf(dmin) and math.isinf(dmax):
                min_pos = point['set']
                max_pos = point['set']
            else:
                min_pos = point['set'] + dmin
                max_pos = point['set'] + dmax

            point['min'] = min_pos
            point['max'] = max_pos

        self[label] = point
        self._update()

    def remove_point(self, label):
        try:
            label = label.upper()
            self.pop(label)
            self._update()
        except Exception as e:
            self.macro.error('Cannot remove label {0}\n{1}'.format(label, e))

    def _update(self):
        try:
            fmt, value = self.json.encode(('', self))
            self.pseudo.getAttribute('configuration').write(value)
            self.macro.debug('Updated configuration:\n{0}'.format(self))
        except Exception as e:
            msg = 'Cannot update configuration]\n{0}\{1}'.format(e, self)
            self.macro.error(msg)

    def __str__(self):
        fmt, value = self.json.encode(('', self), indent=4, sort_keys=True)
        return value


class def_discr_pos(Macro):
    """
    Define a (calibrated) point for a discrete pseudomotor configuration.

    The mandatory parameters to execute the macro are: pseudo, label and pos.

    Two different scenarios exist: To define a new point or to modify an
    existing one. The controller protects from uploading repeated pos values.

    If the point is new, the default dmin and dmax parameters are used to
    construct the calibration. If no set point is provided, the current
    physical position is used instead.

    If the point already exists, the values are updated as in the previous
    case. However, if no dmin and dmax are provided, the previous
    calibration values for dmin and dmax are calculated and used to rebuild
    the calibration.

    .. note::
        The def_discr_pos macro has been included in sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the macro) may occur if
        deemed necessary by the core developers.
    """
    param_def = [
        ['pseudo', Type.PseudoMotor, None, 'Discrete pseudomotor name.'],
        ['label', Type.String, None, 'Label name.'],
        ['pos', Type.Integer, None, 'Discrete (pseudo) position.'],
        ['set', Type.Float, float('inf'), 'Continuous position.'],
        ['dmin', Type.Float, float('-inf'),
         'Delta increment used to define the minimum position.'],
        ['dmax', Type.Float, float('inf'),
         'Delta increment used to define the maximum position.'],
        ]

    def run(self, pseudo, label, pos, setpos, dmin, dmax):
        conf = DiscretePseudoMotorConfiguration(pseudo, self)
        conf.add_point(label, pos, setpos, dmin, dmax)


class udef_discr_pos(Macro):
    """
    Remove a point from a discrete pseudomotor configuration.

    .. note::
        The udef_discr_pos macro has been included in sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the macro) may occur if
        deemed necessary by the core developers.
    """
    param_def = [
        ['pseudo', Type.PseudoMotor, None, 'Discrete pseudomotor name'],
        ['label', Type.String, None, 'Label name'],
    ]

    def run(self, pseudo, label):
        conf = DiscretePseudoMotorConfiguration(pseudo, self)
        conf.remove_point(label)


class prdef_discr(Macro):
    """
    Print discrete pseudomotor configuration.

    .. note::
    The prdef_discr_pos macro has been included in sardana
    on a provisional basis. Backwards incompatible changes
    (up to and including removal of the macro) may occur if
    deemed necessary by the core developers.
    """
    param_def = [
        ['pseudo', Type.PseudoMotor, None, 'Discrete pseudomotor name'],
    ]

    def run(self, pseudo):
        conf = DiscretePseudoMotorConfiguration(pseudo, self)
        col_head_str = [['pos'], ['set'], ['min'], ['max']]
        row_head_str = []
        value_list = []

        for k, v in list(conf.items()):
            row_head_str.append(k)
            _row_values = [k]
            for i in col_head_str:
                _row_values.append(v[i[0]])
            value_list.append(_row_values)

        if len(value_list):
            # Sort by position column
            value_list = sorted(value_list, key=lambda x: x[1])
            # Transpose matrix
            value_list = list(map(list, list(zip(*value_list))))
            # Extract sorted row headers
            row_head_str = value_list[0]
            # Extract sorted values
            value_list = value_list[1:]
            table = Table(value_list, row_head_str=row_head_str,
                          col_head_str=col_head_str, col_head_width=15)
            for line in table.genOutput():
                self.output(line)
        else:
            self.output('No configuration available')



##############################################################################
#
# This file is part of Sardana
#
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from sardana import State
from sardana.pool.controller import CounterTimerController, Type,\
    Description, Access, DataAccess
from taurus import Attribute


class TaurusCounterTimerController(CounterTimerController):
    """
    This controller provides interface to use Taurus Attributes.
    """

    axis_attributes = {
        "TaurusAttribute": {Type: str,
                            Description: 'Taurus attribute ',
                            Access: DataAccess.ReadWrite}
    }

    def __init__(self, inst, props, *args, **kwargs):
        CounterTimerController.__init__(self, inst, props, *args, **kwargs)
        self._axes_taurus_attr = {}

    def AddDevice(self, axis):
        self._axes_taurus_attr[axis] = ''

    def DeleteDevice(self, axis):
        self._axes_taurus_attr.pop(axis)

    def LoadOne(self, axis, value, repetitions, latency):
        pass

    def StateOne(self, axis):
        state = State.On
        status_string = 'The state is ON.'
        return state, status_string

    def StartOne(self, axis, _):
        pass

    def StartAll(self):
        pass

    def ReadOne(self, axis):
        attr = Attribute(self._axes_taurus_attr[axis])
        return attr.read().value

    def AbortOne(self, axis):
        pass

    def getTaurusAttribute(self, axis):
        return self._axes_taurus_attr[axis]

    def setTaurusAttribute(self, axis, value):
        self._axes_taurus_attr[axis] = value



#!/usr/bin/env python

import sys
import time
import PyTango

from upgrade_utils import *


def upgrade(serv, old_vers, new_vers):
    u_kcls = get_suitable_upgrade(old_vers, new_vers)
    u_obj = u_kcls()
    db = PyTango.Database()
    print
    for msg, perc in u_obj.upgrade(db, serv, old_vers, new_vers):
        msg = "\033[1F\033[2K%-70s" % msg
        msg = "%s [ %03d%s ]" % (msg, perc, '%')
        print msg
        time.sleep(0.1)


def main():

    serv = ""
    vers = "0.0.0"

    if len(sys.argv) > 1:
        serv = sys.argv[1]
    if len(sys.argv) > 2:
        vers = sys.argv[2]

    pool_serv_list = get_server_list()

    while not serv in pool_serv_list:
        print_list(pool_serv_list)
        serv = raw_input("Which instance you want to upgrade? ")

    old_vers = get_pool_server_version(serv)

    print "Current version of %s is %s" % (serv, old_vers)

    possible_upgrades = get_possible_upgrades(serv)

    if not possible_upgrades:
        print "Could not find a suitable upgrade plugin for version %s.\nUpgrade FAILED" % old_vers
        return

    while not get_suitable_upgrade(old_vers, vers):
        print_list(possible_upgrades)
        vers = raw_input(
            "To which version you want to upgrade %s (bigger than %s)? " % (serv, old_vers))

    upgrade(serv, old_vers, vers)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from upgrade_utils import *


class Controller:

    def __init__(self, type, file, kcls, name, id):
        self.type = type
        self.file = file
        self.kcls = kcls
        self.name = name
        self.id = id

    def __str__(self):
        return "%s %s %s %s %s" % (self.type, self.file, self.kcls, self.name, self.id)

    def get_controller_prop(self):
        return [self.type, self.file, self.kcls, self.name, self.id]


class Up01To02(Upgrade):

    def get_pool_controllers(self, db, serv):
        """Gets the list of Pool controllers from pool version 0.1.x"""
        pool_serv_name = "Pool/%s" % serv
        pool_dev_name = get_pool_device_from_server(serv, db=db)
        ctrls = db.get_device_property(
            pool_dev_name, ["Controller"])["Controller"]
        i = 0
        ret = []
        while i < len(ctrls):
            type = ctrls[i]
            file = ctrls[i + 1]
            kcls = ctrls[i + 2]
            name = ctrls[i + 3]
            id = ctrls[i + 4]
            i += 5
            ret.append(Controller(type, file, kcls, name, id))
        return ret

    @classmethod
    def fromTo(cls):
        return ("0.1.x", "0.2.0")

    @classmethod
    def supports(cls, old_vers, new_vers):
        return old_vers.startswith('0.1') and new_vers.startswith('0.2')

    @classmethod
    def supports_old(cls, old_vers):
        return old_vers.startswith('0.1')

    @classmethod
    def supports_new(cls, new_vers):
        return new_vers.startswith('0.2')

    def upgrade(self, db, serv, old_vers, new_vers):

        if not Up01To02.supports(old_vers, new_vers):
            raise Exception("Unsupported upgrade")

        yield "Upgrading %s from %s to %s... " % (serv, old_vers, new_vers), 1

        id = 1

        pending_put_properties = {}

        # map used to store <id, name (alias)>
        elem_map, reverse_elem_map = {}, {}

        pool_serv_name = "Pool/%s" % serv
        pool_dev_name = get_pool_device_from_server(serv, db=db)

        # 1 - Update the controller property with unique id
        yield "Updating the controller property with unique id...", 5
        controllers = self.get_pool_controllers(db, serv)

        ctrl_prop_value = []
        for c in controllers:
            c.id = str(id)
            ctrl_prop_value.extend(c.get_controller_prop())
            elem_map[c.id] = c.name.lower()
            reverse_elem_map[c.name.lower()] = c.id
            id += 1

        db.put_device_property(pool_dev_name, {"Controller": ctrl_prop_value})

        # 2 - Update PoolPath: replace ':' by '\n'
        yield "Updating PoolPath: replace ':' by '%sn'..." % '\\', 10
        pool_path = db.get_device_property(
            pool_dev_name, ["PoolPath"])["PoolPath"]

        pool_path_prop_value = []
        for p in pool_path:
            pool_path_prop_value.extend(p.split(":"))
        db.put_device_property(
            pool_dev_name, {"PoolPath": pool_path_prop_value})

        # 3 - Update the properties of all elements
        elems = db.get_device_class_list(pool_serv_name)

        i = 0
        elems_dict = {}
        while i < len(elems):
            dev, kcls = elems[i], elems[i + 1]
            if not elems_dict.has_key(kcls):
                elems_dict[kcls] = []
            elems_dict[kcls].append(dev)
            i += 2

        # 3.1 - For motor remove 'motor_id' and add 'id', 'ctrl_id' and 'axis'
        yield "Updating motors...", 15
        for m_dev_name in elems_dict.get("Motor", []):
            m, ctrl, axis = m_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(m_dev_name, props)
            db.delete_device_property(m_dev_name, ["motor_id"])

            m_alias = db.get_alias(m_dev_name)
            elem_map[id] = m_alias.lower()
            reverse_elem_map[m_alias.lower()] = id

            id += 1

        # 3.2 - For Counter remove 'channel_id' and add 'id', 'ctrl_id' and
        # 'axis'
        yield "Updating counters...", 20
        for expch_dev_name in elems_dict.get("CTExpChannel", []):
            exp, ctrl, axis = expch_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(expch_dev_name, props)
            db.delete_device_property(expch_dev_name, ["channel_id"])

            expch_alias = db.get_alias(expch_dev_name)
            elem_map[id] = expch_alias.lower()
            reverse_elem_map[expch_alias.lower()] = id

            id += 1

        # 3.3 - For 0D remove 'channel_id' and add 'id', 'ctrl_id' and 'axis'
        yield "Updating 0Ds...", 25
        for expch_dev_name in elems_dict.get("ZeroDExpChannel", []):
            exp, ctrl, axis = expch_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(expch_dev_name, props)
            db.delete_device_property(expch_dev_name, ["channel_id"])

            expch_alias = db.get_alias(expch_dev_name)
            elem_map[id] = expch_alias.lower()
            reverse_elem_map[expch_alias.lower()] = id

            id += 1

        # 3.4 - For 1D remove 'channel_id' and add 'id', 'ctrl_id' and 'axis'
        yield "Updating 1Ds...", 30
        for expch_dev_name in elems_dict.get("OneDExpChannel", []):
            exp, ctrl, axis = expch_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(expch_dev_name, props)
            db.delete_device_property(expch_dev_name, ["channel_id"])

            expch_alias = db.get_alias(expch_dev_name)
            elem_map[id] = expch_alias.lower()
            reverse_elem_map[expch_alias.lower()] = id

            id += 1

        # 3.5 - For 2D remove 'channel_id' and add 'id', 'ctrl_id' and 'axis'
        yield "Updating 2Ds...", 35
        for expch_dev_name in elems_dict.get("TwoDExpChannel", []):
            exp, ctrl, axis = expch_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(expch_dev_name, props)
            db.delete_device_property(expch_dev_name, ["channel_id"])

            expch_alias = db.get_alias(expch_dev_name)
            elem_map[id] = expch_alias.lower()
            reverse_elem_map[expch_alias.lower()] = id

            id += 1

        # 3.6 - For CommunicationChannel remove 'channel_id' and add 'id',
        # 'ctrl_id' and 'axis'
        yield "Updating Communication channels...", 40
        for comch_dev_name in elems_dict.get("CommunicationChannel", []):
            com, ctrl, axis = comch_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(comch_dev_name, props)
            db.delete_device_property(comch_dev_name, ["channel_id"])

            comch_alias = db.get_alias(comch_dev_name)
            elem_map[id] = comch_alias.lower()
            reverse_elem_map[comch_alias.lower()] = id

            id += 1

        # 3.7 - For IORegister remove 'channel_id' and add 'id', 'ctrl_id' and
        # 'axis'
        yield "Updating IORegisters...", 45
        for ior_dev_name in elems_dict.get("IORegister", []):
            ior, ctrl, axis = ior_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)], "ctrl_id": [
                str(ctrl)], "axis": [str(axis)]}

            db.put_device_property(ior_dev_name, props)
            db.delete_device_property(ior_dev_name, ["channel_id"])

            ior_alias = db.get_alias(ior_dev_name)
            elem_map[id] = ior_alias.lower()
            reverse_elem_map[ior_alias.lower()] = id

            id += 1

        # 3.8 - For MotorGroup remove 'motor_group_id' and add 'id'
        yield "Updating MotorGroups...", 50
        for mg_dev_name in elems_dict.get("MotorGroup", []):
            mg, inst, name = mg_dev_name.split("/")

            props = {"id": [str(id)]}

            db.put_device_property(mg_dev_name, props)
            db.delete_device_property(
                mg_dev_name, ["motor_group_id", "pool_device"])

            mg_alias = db.get_alias(mg_dev_name)
            elem_map[id] = mg_alias.lower()
            reverse_elem_map[mg_alias.lower()] = id

            # change the format of some properties from list of names to list
            # of IDs
            props = ('motor_list', 'motor_group_list',
                     'pseudo_motor_list', 'phys_group_elt', 'user_group_elt')
            props = db.get_device_property(mg_dev_name, props)

            pending_put_properties[mg_dev_name.lower()] = {}

            new_motor_list = []
            skip = True
            for name in props['motor_list']:
                # if all are already IDs, skip the property
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_motor_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'motor_list': new_motor_list})

            new_motor_group_list = []
            skip = True
            for name in props['motor_group_list']:
                # if all are already IDs, skip the property
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_motor_group_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'motor_group_list': new_motor_group_list})

            new_pseudo_motor_list = []
            skip = True
            for name in props['pseudo_motor_list']:
                # if all are already IDs, skip the property
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_pseudo_motor_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'pseudo_motor_list': new_pseudo_motor_list})

            new_phys_group_elt = []
            skip = True
            for name in props['phys_group_elt']:
                # if all are already IDs, skip the property
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_phys_group_elt.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'phys_group_elt': new_phys_group_elt})

            new_user_group_elt = []
            skip = True
            for name in props['user_group_elt']:
                # if all are already IDs, skip the property
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_user_group_elt.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'user_group_elt': new_user_group_elt})

            id += 1

        # 3.9 - For MeasurementGroup remove 'measurement_group_id' and add 'id'
        yield "Updating MeasurementGroups...", 55
        for mg_dev_name in elems_dict.get("MeasurementGroup", []):
            mg, inst, name = mg_dev_name.split("/")

            props = {"id": [str(id)]}

            db.put_device_property(mg_dev_name, props)
            db.delete_device_property(
                mg_dev_name, ["measurement_group_id",  "pool_device"])

            mg_alias = db.get_alias(mg_dev_name)
            elem_map[id] = mg_alias.lower()
            reverse_elem_map[mg_alias.lower()] = id

            props = ('ct_list', 'zerodexpchannel_list', 'onedexpchannel_list',
                     'twodexpchannel_list', 'pseudocounter_list', 'motor_list',
                     'phys_group_elt', 'user_group_elt')
            props = db.get_device_property(mg_dev_name, props)

            pending_put_properties[mg_dev_name.lower()] = {}

            new_ct_list = []
            skip = True
            for name in props['ct_list']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_ct_list.append(reverse_elem_map.get(
                        name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'ct_list': new_ct_list})

            new_zerodexpchannel_list = []
            skip = True
            for name in props['zerodexpchannel_list']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_zerodexpchannel_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'zerodexpchannel_list': new_zerodexpchannel_list})

            new_onedexpchannel_list = []
            skip = True
            for name in props['onedexpchannel_list']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_onedexpchannel_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'onedexpchannel_list': new_onedexpchannel_list})

            new_twodexpchannel_list = []
            skip = True
            for name in props['twodexpchannel_list']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_twodexpchannel_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'twodexpchannel_list': new_twodexpchannel_list})

            new_pseudocounter_list = []
            skip = True
            for name in props['pseudocounter_list']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_pseudocounter_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'pseudocounter_list': new_pseudocounter_list})

            new_motor_list = []
            skip = True
            for name in props['motor_list']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_motor_list.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'motor_list': new_motor_list})

            new_phys_group_elt = []
            skip = True
            for name in props['phys_group_elt']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_phys_group_elt.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'phys_group_elt': new_phys_group_elt})

            new_user_group_elt = []
            skip = True
            for name in props['user_group_elt']:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    new_user_group_elt.append(
                        reverse_elem_map.get(name.lower(), name.lower()))
            if not skip:
                pending_put_properties[mg_dev_name.lower()].update(
                    {'user_group_elt': new_user_group_elt})

            id += 1

        # 3.10 - For PseudoMotor remove 'pseudo_motor_id', 'motor_group', 'role'
        #        and 'role_idx' and add 'id', 'ctrl_id', 'axis', 'motor_group_id'
        #        Replace motor_list from a list of names to a list of IDs
        yield "Updating PseudoMotors...", 60
        for pm_dev_name in elems_dict.get("PseudoMotor", []):
            pm, ctrl, axis = pm_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)],
                     "ctrl_id": [str(ctrl)],
                     "axis": [str(axis)]}

            prop_values = db.get_device_property(
                pm_dev_name, ["motor_group", "motor_group_id", "motor_list"])
            motor_group = prop_values["motor_group"]

            if not motor_group:
                motor_group_id = motor_group = prop_values["motor_group_id"]
                if not motor_group_id:
                    print "WARNING: neither motor_group nor motor_group_id property is defined for %s." % pm_dev_name
            else:
                motor_group = motor_group[0]
                motor_group_dev_name = db.get_device_alias(motor_group)
                motor_group_id = db.get_device_property(
                    motor_group_dev_name, ["id"])["id"][0]
                props["motor_group_id"] = [str(motor_group_id)]

            db.put_device_property(pm_dev_name, props)
            db.delete_device_property(
                pm_dev_name, ["pseudo_motor_id", "motor_group", "role", "role_idx"])

            skip = True
            motor_ids = []
            for name in prop_values["motor_list"]:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    motor_ids.append(reverse_elem_map[name.lower()])
            if not skip:
                db.put_device_property(pm_dev_name, {"motor_list": motor_ids})

            pm_alias = db.get_alias(pm_dev_name)
            elem_map[id] = pm_alias.lower()
            reverse_elem_map[pm_alias.lower()] = id

            id += 1

        # 3.11 - For PseudoCounter remove 'pseudo_counter_id', 'role' and
        #        'role_idx' and add 'id', 'ctrl_id', 'axis'
        #        Replace channel_list from a list of names to a list of IDs
        yield "Updating PseudoCounters...", 65
        for pc_dev_name in elems_dict.get("PseudoCounter", []):
            pc, ctrl, axis = pc_dev_name.split("/")

            for c in controllers:
                if c.name.lower() == ctrl.lower():
                    ctrl = c.id

            props = {"id": [str(id)],
                     "ctrl_id": [str(ctrl)],
                     "axis": [str(axis)]}

            db.put_device_property(pc_dev_name, props)
            db.delete_device_property(
                pc_dev_name, ["pseudo_counter_id", "role", "role_idx"])

            channel_list = db.get_device_property(
                pc_dev_name, "channel_list")["channel_list"]

            skip = True
            channel_ids = []
            for name in channel_list:
                try:
                    int(name)
                except Exception, e:
                    skip = False
                    channel_ids.append(reverse_elem_map[name.lower()])
            if not skip:
                db.put_device_property(
                    pm_dev_name, {"channel_list": channel_ids})

            pc_alias = db.get_alias(pc_dev_name)
            elem_map[id] = pc_alias.lower()
            reverse_elem_map[pc_alias.lower()] = id

            id += 1

        # 4. - Apply pending properties
        yield "Applying pending properties...", 75
        for dev, props in pending_put_properties.iteritems():
            for name, val in props.iteritems():
                for i, elem in enumerate(val):
                    if isinstance(elem, str):
                        val[i] = reverse_elem_map[elem]
            db.put_device_property(dev, props)

        # 5. - Finally update the version property in the database
        yield "Updating the version property in the database...", 85
        db.put_device_property(pool_dev_name, {"Version": [new_vers]})

        # 6. - Check if the service is already registered
        yield "Checking if the service is already registered...", 95
        services = db.get_services("Sardana", serv)
        if len(services) == 0:
            db.register_service("Sardana", serv, pool_dev_name)

        yield "Finished", 100



##########################################################################
##
# This file is part of PyTango, a python binding for Tango
##
# http://www.tango-controls.org/static/PyTango/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# PyTango is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# PyTango is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with PyTango.  If not, see <http://www.gnu.org/licenses/>.
##
##########################################################################

"""reST directive for syntax-highlighting ipython interactive sessions.

XXX - See what improvements can be made based on the new (as of Sept 2009)
'pycon' lexer for the python console.  At the very least it will give better
highlighted tracebacks.
"""

#-----------------------------------------------------------------------------
# Needed modules

# Standard library
import re

# Third party
from pygments.lexer import Lexer, do_insertions
from pygments.lexers.agile import (PythonConsoleLexer, PythonLexer,
                                   PythonTracebackLexer)
from pygments.token import Comment, Generic

from sphinx import highlighting

#-----------------------------------------------------------------------------
# Global constants
line_re = re.compile('.*?\n')

#-----------------------------------------------------------------------------
# Code begins - classes and functions


class IPythonConsoleLexer(Lexer):
    """
    For IPython console output or doctests, such as:

    .. sourcecode:: ipython

      In [1]: a = 'foo'

      In [2]: a
      Out[2]: 'foo'

      In [3]: print(a)
      foo

      In [4]: 1 / 0

    Notes:

      - Tracebacks are not currently supported.

      - It assumes the default IPython prompts, not customized ones.
    """

    name = 'IPython console session'
    aliases = ['ipython']
    mimetypes = ['text/x-ipython-console']
    input_prompt = re.compile("(In \[(?P<N>[0-9]+)\]: )|(   \.\.\.+:)")
    output_prompt = re.compile("(Out\[(?P<N>[0-9]+)\]: )|(   \.\.\.+:)")
    continue_prompt = re.compile("   \.\.\.+:")
    tb_start = re.compile("\-+")

    def get_tokens_unprocessed(self, text):
        pylexer = PythonLexer(**self.options)
        tblexer = PythonTracebackLexer(**self.options)

        curcode = ''
        insertions = []
        for match in line_re.finditer(text):
            line = match.group()
            input_prompt = self.input_prompt.match(line)
            continue_prompt = self.continue_prompt.match(line.rstrip())
            output_prompt = self.output_prompt.match(line)
            if line.startswith("#"):
                insertions.append((len(curcode),
                                   [(0, Comment, line)]))
            elif input_prompt is not None:
                insertions.append((len(curcode),
                                   [(0, Generic.Prompt, input_prompt.group())]))
                curcode += line[input_prompt.end():]
            elif continue_prompt is not None:
                insertions.append((len(curcode),
                                   [(0, Generic.Prompt, continue_prompt.group())]))
                curcode += line[continue_prompt.end():]
            elif output_prompt is not None:
                # Use the 'error' token for output.  We should probably make
                # our own token, but error is typicaly in a bright color like
                # red, so it works fine for our output prompts.
                insertions.append((len(curcode),
                                   [(0, Generic.Error, output_prompt.group())]))
                curcode += line[output_prompt.end():]
            else:
                if curcode:
                    for item in do_insertions(insertions,
                                              pylexer.get_tokens_unprocessed(curcode)):
                        yield item
                        curcode = ''
                        insertions = []
                yield match.start(), Generic.Output, line
        if curcode:
            for item in do_insertions(insertions,
                                      pylexer.get_tokens_unprocessed(curcode)):
                yield item


def setup(app):
    """Setup as a sphinx extension."""

    # This is only a lexer, so adding it below to pygments appears sufficient.
    # But if somebody knows that the right API usage should be to do that via
    # sphinx, by all means fix it here.  At least having this setup.py
    # suppresses the sphinx warning we'd get without it.
    pass

#-----------------------------------------------------------------------------
# Register the extension as a valid pygments lexer
highlighting.lexers['ipython'] = IPythonConsoleLexer()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import sys

import sardana
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.base import TaurusBaseWidget

from taurus.external.qt import QtCore, QtGui

import taurus.core
from taurus.qt.qtcore.communication import SharedDataManager
from taurus.qt.qtgui.input import TaurusValueLineEdit

from .displayscanangles import DisplayScanAngles

import taurus.core.util.argparse
import taurus.qt.qtgui.application
from taurus.qt.qtgui.util.ui import UILoadable

from PyTango import *
from sardana.taurus.qt.qtgui.extra_macroexecutor import TaurusMacroExecutorWidget, TaurusSequencerWidget, \
    TaurusMacroConfigurationDialog, \
    TaurusMacroDescriptionViewer, DoorOutput, DoorDebug, DoorResult

__docformat__ = 'restructuredtext'


class EngineModesComboBox(Qt.QComboBox, TaurusBaseWidget):
    """ComboBox representing list of engine modes"""

    def __init__(self, parent=None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TaurusBaseWidget, name)
        self.setSizeAdjustPolicy(Qt.QComboBox.AdjustToContentsOnFirstShow)
        self.setToolTip("Choose a engine mode ...")
        QtCore.QMetaObject.connectSlotsByName(self)

    def loadEngineModeNames(self, enginemodes):
        self.clear()
        self.addItems(enginemodes)


@UILoadable(with_ui="_ui")
class HKLScan(TaurusWidget):

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode=designMode)

        self.loadUi(filename="hklscan.ui")

        self._ui.hklStartScanButton.clicked.connect(self.start_hklscan)
        self._ui.hklStopScanButton.clicked.connect(self.stop_hklscan)
        self._ui.hklDisplayAnglesButton.clicked.connect(self.display_angles)
        self._ui.MacroServerConnectionButton.clicked.connect(
            self.open_macroserver_connection_panel)

        # Create a global SharedDataManager
        Qt.qApp.SDM = SharedDataManager(self)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'hklscan'
        ret['group'] = 'Taurus Containers'
        ret['container'] = ':/designer/frame.png'
        ret['container'] = True
        return ret

    def setModel(self, model):
        if model is not None:
            self.device = taurus.Device(model)

        self.pseudo_motor_names = []
        for motor in self.device.hklpseudomotorlist:
            self.pseudo_motor_names.append(motor.split(' ')[0])

        self.h_device_name = self.pseudo_motor_names[0]
        self.h_device = taurus.Device(self.h_device_name)
        self.k_device_name = self.pseudo_motor_names[1]
        self.k_device = taurus.Device(self.k_device_name)
        self.l_device_name = self.pseudo_motor_names[2]
        self.l_device = taurus.Device(self.l_device_name)

        # Add dynamically the angle widgets

        motor_list = self.device.motorlist
        motor_names = []
        for motor in self.device.motorlist:
            motor_names.append(motor.split(' ')[0])

        self.nb_motors = len(motor_list)

        angles_labels = []
        angles_names = []
        angles_taurus_label = []

        gap_x = 800 // self.nb_motors

        try:
            angles_names = self.device.motorroles
        except:  # Only for compatibility
            if self.nb_motors == 4:
                angles_names.append("omega")
                angles_names.append("chi")
                angles_names.append("phi")
                angles_names.append("theta")
            elif self.nb_motors == 6:
                angles_names.append("mu")
                angles_names.append("th")
                angles_names.append("chi")
                angles_names.append("phi")
                angles_names.append("gamma")
                angles_names.append("delta")

        for i in range(0, self.nb_motors):
            angles_labels.append(QtGui.QLabel(self))
            angles_labels[i].setGeometry(
                QtCore.QRect(50 + gap_x * i, 290, 51, 17))
            alname = "angleslabel" + str(i)
            angles_labels[i].setObjectName(alname)
            angles_labels[i].setText(QtGui.QApplication.translate(
                "HKLScan", angles_names[i], None))
            angles_taurus_label.append(TaurusLabel(self))
            angles_taurus_label[i].setGeometry(
                QtCore.QRect(50 + gap_x * i, 320, 81, 19))
            atlname = "anglestauruslabel" + str(i)
            angles_taurus_label[i].setObjectName(atlname)
            angles_taurus_label[i].setModel(motor_names[i] + "/Position")

        # Set model to hkl display

        hmodel = self.h_device_name + "/Position"
        self._ui.taurusValueLineH.setModel(hmodel)
        self._ui.taurusLabelValueH.setModel(hmodel)
        kmodel = self.k_device_name + "/Position"
        self._ui.taurusValueLineK.setModel(kmodel)
        self._ui.taurusLabelValueK.setModel(kmodel)
        lmodel = self.l_device_name + "/Position"
        self._ui.taurusValueLineL.setModel(lmodel)
        self._ui.taurusLabelValueL.setModel(lmodel)

        # Set model to engine and modes

        enginemodel = model + '/engine'
        self._ui.taurusLabelEngine.setModel(enginemodel)
        enginemodemodel = model + '/enginemode'
        self._ui.taurusLabelEngineMode.setModel(enginemodemodel)

        self.enginemodescombobox = EngineModesComboBox(self)
        self.enginemodescombobox.setGeometry(QtCore.QRect(150, 445, 221, 27))
        self.enginemodescombobox.setObjectName("enginemodeslist")

        self.enginemodescombobox.loadEngineModeNames(self.device.hklmodelist)

        self.enginemodescombobox.currentIndexChanged['QString'].connect(
            self.onModeChanged)

    @Qt.pyqtSlot('QString')
    def onModeChanged(self, modename):
        if self.device.engine != "hkl":
            self.device.write_attribute("engine", "hkl")
        self.device.write_attribute("enginemode", str(modename))

    def start_hklscan(self):
        start_hkl = []
        stop_hkl = []
        start_hkl.append(float(self._ui.lineEditStartH.text()))
        start_hkl.append(float(self._ui.lineEditStartK.text()))
        start_hkl.append(float(self._ui.lineEditStartL.text()))
        stop_hkl.append(float(self._ui.lineEditStopH.text()))
        stop_hkl.append(float(self._ui.lineEditStopK.text()))
        stop_hkl.append(float(self._ui.lineEditStopL.text()))
        nb_points = int(self._ui.LineEditNbpoints.text())
        sample_time = float(self._ui.LineEditSampleTime.text())
        dim = 0
        macro_name = ["ascan", "a2scan", "a3scan"]
        macro_command = []
        index_to_scan = []
        if self.door_device is not None:
            for i in range(0, 3):
                if start_hkl[i] != stop_hkl[i]:
                    dim = dim + 1
                    index_to_scan.append(i)
            if dim > 0:
                macro_command.append(macro_name[dim - 1])
                for i in range(len(index_to_scan)):
                    macro_command.append(
                        str(self.pseudo_motor_names[index_to_scan[i]]))
                    macro_command.append(str(start_hkl[index_to_scan[i]]))
                    macro_command.append(str(stop_hkl[index_to_scan[i]]))
                macro_command.append(str(nb_points))
                macro_command.append(str(sample_time))
                self.door_device.RunMacro(macro_command)

    def stop_hklscan(self):
        self.door_device.StopMacro()

    def display_angles(self):

        xangle = []
        for i in range(0, 6):
            xangle.append(40 + i * 100)

        yhkl = 50

        tr = self.device.selectedtrajectory

        w = DisplayScanAngles()

        angles_labels = []
        angles_names = []

        if self.nb_motors == 4:
            angles_names.append("omega")
            angles_names.append("chi")
            angles_names.append("phi")
            angles_names.append("theta")
        elif self.nb_motors == 6:
            angles_names.append("mu")
            angles_names.append("th")
            angles_names.append("chi")
            angles_names.append("phi")
            angles_names.append("gamma")
            angles_names.append("delta")

        dsa_label = []
        for i in range(0, self.nb_motors):
            dsa_label.append(QtGui.QLabel(w))
            dsa_label[i].setGeometry(QtCore.QRect(xangle[i], yhkl, 51, 20))
            label_name = "dsa_label_" + str(i)
            dsa_label[i].setObjectName(label_name)
            dsa_label[i].setText(QtGui.QApplication.translate(
                "Form", angles_names[i], None))

        start_hkl = []
        stop_hkl = []
        missed_values = 0
        # TODO: This code will raise exception if one of the line edits is empty.
        # But not all dimensions (H & K & L) are obligatory. One could try
        # to display angles of just 1 or 2 dimensional scan.
        try:
            start_hkl.append(float(self._ui.lineEditStartH.text()))
            start_hkl.append(float(self._ui.lineEditStartK.text()))
            start_hkl.append(float(self._ui.lineEditStartL.text()))
            stop_hkl.append(float(self._ui.lineEditStopH.text()))
            stop_hkl.append(float(self._ui.lineEditStopK.text()))
            stop_hkl.append(float(self._ui.lineEditStopL.text()))
            nb_points = int(self._ui.LineEditNbpoints.text())
        except:
            nb_points = -1
            missed_values = 1

        increment_hkl = []

        if nb_points > 0:
            for i in range(0, 3):
                increment_hkl.append((stop_hkl[i] - start_hkl[i]) / nb_points)

        taurusValueAngle = []

        for i in range(0, nb_points + 1):
            hkl_temp = []
            for j in range(0, 3):
                hkl_temp.append(start_hkl[j] + i * increment_hkl[j])

            no_trajectories = 0
            try:
                self.device.write_attribute("computetrajectoriessim", hkl_temp)
            except:
                no_trajectories = 1

            if not no_trajectories:

                angles_list = self.device.trajectorylist[tr]

                taurusValueAngle.append([])

                for iangle in range(0, self.nb_motors):
                    taurusValueAngle[i].append(TaurusValueLineEdit(w))
                    taurusValueAngle[i][iangle].setGeometry(
                        QtCore.QRect(xangle[iangle], yhkl + 30 * (i + 1), 80, 27))
                    taurusValueAngle[i][iangle].setReadOnly(True)
                    tva_name = "taurusValueAngle" + str(i) + "_" + str(iangle)
                    taurusValueAngle[i][iangle].setObjectName(tva_name)
                    taurusValueAngle[i][iangle].setValue(
                        "%10.4f" % angles_list[iangle])
            else:
                taurusValueAngle.append(TaurusValueLineEdit(w))
                taurusValueAngle[i].setGeometry(QtCore.QRect(
                    xangle[0], yhkl + 30 * (i + 1), self.nb_motors * 120, 27))
                taurusValueAngle[i].setReadOnly(True)
                tva_name = "taurusValueAngle" + str(i)
                taurusValueAngle[i].setObjectName(tva_name)
                taurusValueAngle[i].setValue(
                    "...                             No angle solution for hkl values                             ...")
        # TODO: not all dimensions (H & K & L) are obligatory. One could try
        # to display angles of just 1 or 2 dimensional scan.
        if nb_points == -1:
            nb_points = 0
            taurusValueAngle.append(TaurusValueLineEdit(w))
            taurusValueAngle[0].setGeometry(QtCore.QRect(
                xangle[0], yhkl + 30, self.nb_motors * 120, 27))
            taurusValueAngle[0].setReadOnly(True)
            tva_name = "taurusValueAngle"
            taurusValueAngle[0].setObjectName(tva_name)
            taurusValueAngle[0].setValue(
                "...          No scan parameters filled. Fill them in the main window         ...")

        w.resize(self.nb_motors * 140, 120 + nb_points * 40)

        w.show()
        w.show()

    def open_macroserver_connection_panel(self):
        w = TaurusMacroConfigurationDialog(self)
        Qt.qApp.SDM.connectReader("macroserverName", w.selectMacroServer)
        Qt.qApp.SDM.connectReader("doorName", w.selectDoor)
        Qt.qApp.SDM.connectReader("doorName", self.onDoorChanged)
        Qt.qApp.SDM.connectWriter(
            "macroserverName", w, 'macroserverNameChanged')
        Qt.qApp.SDM.connectWriter("doorName", w, 'doorNameChanged')

        w.show()

    def onDoorChanged(self, doorName):
        if doorName != self.door_device_name:
            self.door_device_name = doorName
            self.door_device = taurus.Device(doorName)


def main():

    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.usage = "%prog  <model> [door_name]"
    parser.set_description("a taurus application for performing hkl scans")

    app = taurus.qt.qtgui.application.TaurusApplication(
        cmd_line_parser=parser, app_version=sardana.Release.version)
    app.setApplicationName("hklscan")
    args = app.get_command_line_args()
    if len(args) < 1:
        msg = "model not set (requires diffractometer controller)"
        parser.error(msg)

    w = HKLScan()
    w.model = args[0]
    w.setModel(w.model)

    w.door_device = None
    w.door_device_name = None
    if len(args) > 1:
        w.onDoorChanged(args[1])
    else:
        print("WARNING: Not door name supplied. Connection to "
              "MacroServer/Door not automatically done")
    w.show()

    sys.exit(app.exec_())

 #   if len(sys.argv)>1: model=sys.argv[1]
 #   else: model = None
 #   app = Qt.QApplication(sys.argv)
 #   w = HKLScan()
 #   w.setModel(model)
 #   w.show()
 #   sys.exit(app.exec_())

if __name__ == "__main__":
    main()



import PyTango
import CounterTimerController

nb_call = 0


class Vct6Controller(CounterTimerController.CounterTimerController):
    "This class is the Tango Sardana CounterTimer controller for the VCT6"

    ctrl_extra_attributes = {'PyCT_extra_1': {'Type': 'PyTango.DevDouble', 'R/W Type': 'PyTango.READ_WRITE'},
                             'PyCT_extra_2': {'Type': 'PyTango.DevLong', 'R/W Type': 'PyTango.READ'},
                             'PyCT_extra_3': {'Type': 'PyTango.DevBoolean', 'R/W Type': 'PyTango.READ'}}

    class_prop = {'CtrlDevName': {'Type': 'PyTango.DevString',
                                  'Description': 'The ctrl simulator Tango device name'}}

    MaxDevice = 6

    def __init__(self, inst, props):
        CounterTimerController.CounterTimerController.__init__(
            self, inst, props)
        print "PYTHON -> Vct6Controller ctor for instance", inst

        global nb_call

        if nb_call <= 1:
            nb_call = nb_call + 1
            raise NameError, "Ouuups"

        self.ct_name = "Vct6Controller/" + self.inst_name
        self.simu_ctrl = None
        self.simu_ctrl = PyTango.DeviceProxy(self.CtrlDevName)

        try:
            self.simu_ctrl.ping()
        except:
            self.simu_ctrl = None
            raise

    def AddDevice(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In AddDevice method for index", ind
#        raise RuntimeError,"Hola la la"

    def DeleteDevice(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In DeleteDevice method for index", ind

    def StateOne(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In StateOne method for index", ind
        if self.simu_ctrl is not None:
            sta = self.simu_ctrl.command_inout("GetCounterState", ind)
            print "State in controller =", sta
            tup = (sta, "Status error string from controller")
        else:
            raise RuntimeError, "Ctrl Tango's proxy null!!!"
        return tup

    def PreReadAll(self):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In PreReadAll method"

    def PreReadOne(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In PreReadOne method for index", ind

    def ReadAll(self):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In ReadAll method"

    def ReadOne(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In ReadOne method for index", ind
        return 123

    def GetExtraAttributePar(self, ind, name):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In GetExtraFeaturePar method for index", ind, " name=", name
        if name == "PyCT_extra_1":
            return 88.99

        if name == "PyCT_extra_2":
            return 33
        return 2233

    def SetExtraAttributePar(self, ind, name, value):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In SetExtraFeaturePar method for index", ind, " name=", name, " value=", value

    def SendToCtrl(self, in_data):
        print "Received value =", in_data
        return "Adios"

    def StartOneCT(self, ind):
        print "In StartOneCT"

    def LoadAll():
        print "In LoadAll"

    def __del__(self):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": Aarrrrrg, I am dying"


if __name__ == "__main__":
    obj = Vct6Controller('test')
#    obj.AddDevice(2)
#    obj.DeleteDevice(2)



##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from sardana import State
from sardana.util.funcgenerator import FunctionGenerator
from sardana.pool.controller import TriggerGateController
from sardana.sardanathreadpool import get_thread_pool
from sardana.pool.pooldefs import SynchDomain


class DummyTriggerGateController(TriggerGateController):
    """Basic controller intended for demonstration purposes only.
    """
    gender = "Simulation"
    organization = "ALBA-Cells"
    MaxDevice = 1

    def __init__(self, inst, props, *args, **kwargs):
        """Constructor"""
        TriggerGateController.__init__(self, inst, props, *args, **kwargs)
        self.tg = {}
        self.conf = {}

    def SynchOne(self, axis, conf):
        idx = axis - 1
        tg = self.tg[idx]
        tg.set_configuration(conf)
        self.conf[idx] = conf

    def AddDevice(self, axis):
        self._log.debug('AddDevice(%d): entering...' % axis)
        idx = axis - 1
        func_generator = FunctionGenerator()
        func_generator.initial_domain = SynchDomain.Time
        func_generator.active_domain = SynchDomain.Time
        self.tg[idx] = func_generator

    def StateOne(self, axis):
        """Get the dummy trigger/gate state"""
        try:
            self._log.debug('StateOne(%d): entering...' % axis)
            sta = State.On
            status = "Stopped"
            idx = axis - 1
            tg = self.tg[idx]
            if tg.is_running() or tg.is_started():
                sta = State.Moving
                status = "Moving"
            self._log.debug('StateOne(%d): returning (%s, %s)' %
                            (axis, sta, status))
        except Exception as e:
            print(e)
        return sta, status

    def PreStartAll(self):
        pass

    def StartAll(self):
        pass

    def PreStartOne(self, axis, value=None):
        return True

    def StartOne(self, axis):
        """Start the specified trigger
        """
        self._log.debug('StartOne(%d): entering...' % axis)
        idx = axis - 1
        tg = self.tg[idx]
        tg.start()
        get_thread_pool().add(tg.run)

    def AbortOne(self, axis):
        """Start the specified trigger
        """
        self._log.debug('AbortOne(%d): entering...' % axis)
        idx = axis - 1
        self.tg[idx].stop()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package contains macro parameters parsing utilities."""

__all__ = ["ParamParser"]

import re
import collections

# Token specification
PARAM = r"(?P<PARAM>[^\[\]\s]+)"
QUOTEDPARAM = r'"(?P<QUOTEDPARAM>.*?)(?<!\\)"'
SINGQUOTEDPARAM = r"'(?P<SINGQUOTEDPARAM>.*?)'"
LPAREN = r"(?P<LPAREN>\[)"
RPAREN = r"(?P<RPAREN>\])"
WS = r"(?P<WS>\s+)"

master_pat = re.compile("|".join([QUOTEDPARAM, SINGQUOTEDPARAM, PARAM, LPAREN,
                                  RPAREN, WS]))

# Tokenizer
Token = collections.namedtuple("Token", ["type", "value"])


def generate_tokens(text):
    scanner = master_pat.scanner(text)
    for m in iter(scanner.match, None):
        # quoted parameters must be returned without the quotes that's why we
        # extract a given group, otherwise we would extract the whole match
        group = (m.group("QUOTEDPARAM")
                 or m.group("SINGQUOTEDPARAM")
                 or m.group())
        tok = Token(m.lastgroup, group)
        if tok.type != "WS":
            yield tok


def is_repeat_param(param_def):
    return isinstance(param_def["type"], list)


def is_repeat_param_single(param_def):
    return len(param_def) == 1


class ParseError(Exception):
    pass


class UnrecognizedParamValue(ParseError):
    pass


class ExcessParamValue(ParseError):
    pass


class ParamParser:
    """Implementation of a recursive descent parser. Use the ._accept() method
    to test and accept the current lookahead token. Use the ._expect()
    method to exactly match and discard the next token on the input
    (or raise a SyntaxError if it doesn't match).

    Inspired on Python Cookbook 3 (chapter 2.19)
    """

    def __init__(self, params_def=None):
        self._params_def = params_def

    def parse(self, text):
        self.tokens = generate_tokens(text)
        self.tok = None             # Last symbol consumed
        self.nexttok = None         # Next symbol tokenized
        self._advance()             # Load first lookahead token
        params = self._params()
        self._end_check()
        return params

    def _advance(self):
        """Advance one token ahead"""
        self.tok, self.nexttok = self.nexttok, next(self.tokens, None)

    def _accept(self, toktype):
        """Test and consume the next token if it matches toktype"""
        if self.nexttok and self.nexttok.type == toktype:
            self._advance()
            return True
        else:
            return False

    def _expect(self, toktype):
        """Consume next token if it matches toktype or raise SyntaxError"""
        if not self._accept(toktype):
            raise ParseError("Expected " + toktype)

    # Grammar rules follow

    def _params(self, params_def=None, is_repeat=False):
        """Interpret parameter values by iterating over generated tokens
        according to parameters definition.

        It is used either at the macro level or a the repeat parameter
        repetition level.

        :param params_def: parameters definition as used by the
            :meth:`sardana.macroserver.msmetamacro.Parametrizable.get_parameter`
            or by the
            `attr:`sardana.taurus.core.tango.sardana.macro.MacroInfo.parameters`
        :type params_def: list<dict>
        :param end_check: whether to check if there are parameter values
            exceeding parameters definition
        :type end_check: bool
        :return: parameter values
        :rtype: list
        """
        params_def = params_def or self._params_def
        len_params_def = len(params_def)
        params = []
        for param_idx, param_def in enumerate(params_def):
            # no next tokens means that the string being parsed had finished
            if self.nexttok is None:
                break
            if is_repeat_param(param_def):
                is_last_param = False
                if param_idx == len_params_def - 1:
                    is_last_param = True
                repeat_param_def = param_def["type"]
                param_value = self._repeat_param(repeat_param_def,
                                                 is_last_param)
            else:
                try:
                    param_value = self._param()
                except UnrecognizedParamValue:
                    # this exception may occur if repeat is not complete -
                    # uses default values
                    if is_repeat:
                        return params
                    raise
            params.append(param_value)
        return params

    def _param(self):
        """Interpret normal parameter value. Respect quotes for string
        parameters.

        :return: parameter value
        :rtype: str
        """
        if self._accept("LPAREN"):
            # empty brackets will be interpreted as a default value
            self._expect("RPAREN")
            param = []
        elif self._accept("QUOTEDPARAM"):
            # quoted parameters allows using quotes escaped by \\
            string = self.tok.value
            string = string.replace('\\"', '"')
            param = string
        elif self._accept("SINGQUOTEDPARAM"):
            param = self.tok.value
        elif self._accept("PARAM"):
            tok_value = self.tok.value
            param = tok_value
        else:
            msg = "%s is not a valid param value" % self.tok.value
            raise UnrecognizedParamValue(msg)
        return param

    def _repeat_param(self, repeat_param_def, is_last_param):
        """Interpret repeat parameter.

        Accepts repeat parameters using the following rules:
        * enclosed in parenthesis
        * non-enclosed in parenthesis multiple repetitions of the last repeat
          parameter (can be single or multiple)
        * non-enclosed in parenthesis one repetition of single repeat
        parameter at arbitrary position

        :param repeat_param_def: repeat parameter definition
        :type repeat_param_def: list<dict>
        :param is_last_param: whether this repeat parameter is the last in the
            definition
        :type is_last_param: bool
        :return: repeat parameter value
        :rtype: list
        """
        repeats = []

        if self._accept("LPAREN"):
            while True:
                repeat = self._repeat(repeat_param_def)
                if repeat is None:
                    break
                repeats.append(repeat)
            self._expect("RPAREN")
        else:
            single = is_repeat_param_single(repeat_param_def)
            if is_last_param:
                while True:
                    repeat = []
                    for _ in repeat_param_def:
                        try:
                            param = self._param()
                        except UnrecognizedParamValue:
                            return repeats
                        if single:
                            repeat = param
                        else:
                            repeat.append(param)
                    repeats.append(repeat)
            elif single:
                param = self._param()
                repeats = [param]
        return repeats

    def _repeat(self, repeat_param_def):
        """Interpret one repetition of the repeat parameter.

        :param repeat_param_def: repeat parameter definition
        :type repeat_param_def: list<dict>
        :return: repeat value
        :rtype: list or None
        """
        repeat = None
        if self._accept("LPAREN"):
            # empty brackets will be interpreted as a default value
            if self._accept("RPAREN"):
                repeat = []
            else:
                repeat = self._params(repeat_param_def, is_repeat=True)
                # repetitions of single repeat parameters are not enclosed
                # in parenthesis so remove it
                if is_repeat_param_single(repeat_param_def):
                    repeat = repeat[0]
                self._expect("RPAREN")
        else:
            try:
                repeat = self._param()
            except UnrecognizedParamValue:
                # no repeat found - return None
                pass
        return repeat

    def _end_check(self):
        """Check if there are excessive tokens."""
        excess_tokens = ""
        if len(self._params_def) == 0 and self.nexttok is not None:
            excess_tokens += self.nexttok.value
        while True:
            self._advance()
            if self.nexttok is None:
                break
            excess_tokens += self.nexttok.value
        if len(excess_tokens) > 0:
            raise ExcessParamValue("excess tokens are %s" % excess_tokens)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'poolEditor.ui'
#
# Created: Wed Nov 10 10:49:04 2010
#      by: PyQt4 UI code generator 4.4.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui


class Ui_PoolEditor(object):

    def setupUi(self, PoolEditor):
        PoolEditor.setObjectName("PoolEditor")
        PoolEditor.resize(400, 333)
        self.gridLayout = QtGui.QGridLayout(PoolEditor)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout_2 = QtGui.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.instanceNameLineEdit = QtGui.QLineEdit(PoolEditor)
        self.instanceNameLineEdit.setObjectName("instanceNameLineEdit")
        self.gridLayout_2.addWidget(self.instanceNameLineEdit, 0, 1, 1, 1)
        self.instanceNameLabel = QtGui.QLabel(PoolEditor)
        self.instanceNameLabel.setObjectName("instanceNameLabel")
        self.gridLayout_2.addWidget(self.instanceNameLabel, 0, 0, 1, 1)
        self.poolDeviceNameLineEdit = QtGui.QLineEdit(PoolEditor)
        self.poolDeviceNameLineEdit.setObjectName("poolDeviceNameLineEdit")
        self.gridLayout_2.addWidget(self.poolDeviceNameLineEdit, 1, 1, 1, 1)
        self.aliasLineEdit = QtGui.QLineEdit(PoolEditor)
        self.aliasLineEdit.setObjectName("aliasLineEdit")
        self.gridLayout_2.addWidget(self.aliasLineEdit, 2, 1, 1, 1)
        self.poolVersionLineEdit = QtGui.QLineEdit(PoolEditor)
        self.poolVersionLineEdit.setObjectName("poolVersionLineEdit")
        self.gridLayout_2.addWidget(self.poolVersionLineEdit, 3, 1, 1, 1)
        self.poolDeviceNameLabel = QtGui.QLabel(PoolEditor)
        self.poolDeviceNameLabel.setObjectName("poolDeviceNameLabel")
        self.gridLayout_2.addWidget(self.poolDeviceNameLabel, 1, 0, 1, 1)
        self.aliasLabel = QtGui.QLabel(PoolEditor)
        self.aliasLabel.setObjectName("aliasLabel")
        self.gridLayout_2.addWidget(self.aliasLabel, 2, 0, 1, 1)
        self.poolVersionLabel = QtGui.QLabel(PoolEditor)
        self.poolVersionLabel.setObjectName("poolVersionLabel")
        self.gridLayout_2.addWidget(self.poolVersionLabel, 3, 0, 1, 1)
        self.poolDeviceNameCheckBox = QtGui.QCheckBox(PoolEditor)
        self.poolDeviceNameCheckBox.setObjectName("poolDeviceNameCheckBox")
        self.gridLayout_2.addWidget(self.poolDeviceNameCheckBox, 1, 2, 1, 1)
        self.aliasCheckBox = QtGui.QCheckBox(PoolEditor)
        self.aliasCheckBox.setObjectName("aliasCheckBox")
        self.gridLayout_2.addWidget(self.aliasCheckBox, 2, 2, 1, 1)
        self.poolVersionCheckBox = QtGui.QCheckBox(PoolEditor)
        self.poolVersionCheckBox.setObjectName("poolVersionCheckBox")
        self.gridLayout_2.addWidget(self.poolVersionCheckBox, 3, 2, 1, 1)
        self.gridLayout.addLayout(self.gridLayout_2, 0, 0, 1, 1)
        self.groupBox = QtGui.QGroupBox(PoolEditor)
        self.groupBox.setObjectName("groupBox")
        self.horizontalLayout = QtGui.QHBoxLayout(self.groupBox)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.poolPathList = QtGui.QListWidget(self.groupBox)
        self.poolPathList.setObjectName("poolPathList")
        self.horizontalLayout.addWidget(self.poolPathList)
        self.verticalLayout = QtGui.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.addButton = QtGui.QToolButton(self.groupBox)
        self.addButton.setObjectName("addButton")
        self.verticalLayout.addWidget(self.addButton)
        self.removeButton = QtGui.QToolButton(self.groupBox)
        self.removeButton.setObjectName("removeButton")
        self.verticalLayout.addWidget(self.removeButton)
        self.upButton = QtGui.QToolButton(self.groupBox)
        self.upButton.setArrowType(QtCore.Qt.NoArrow)
        self.upButton.setObjectName("upButton")
        self.verticalLayout.addWidget(self.upButton)
        self.downButton = QtGui.QToolButton(self.groupBox)
        self.downButton.setObjectName("downButton")
        self.verticalLayout.addWidget(self.downButton)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.gridLayout.addWidget(self.groupBox, 1, 0, 1, 1)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtGui.QSpacerItem(
            40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.createButton = QtGui.QPushButton(PoolEditor)
        self.createButton.setObjectName("createButton")
        self.horizontalLayout_2.addWidget(self.createButton)
        self.closeButton = QtGui.QPushButton(PoolEditor)
        self.closeButton.setObjectName("closeButton")
        self.horizontalLayout_2.addWidget(self.closeButton)
        self.gridLayout.addLayout(self.horizontalLayout_2, 2, 0, 1, 1)

        self.retranslateUi(PoolEditor)
        QtCore.QObject.connect(self.poolDeviceNameCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self.poolDeviceNameLineEdit.setDisabled)
        QtCore.QObject.connect(self.aliasCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self.aliasLineEdit.setDisabled)
        QtCore.QObject.connect(self.poolVersionCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self.poolVersionLineEdit.setDisabled)
        QtCore.QObject.connect(self.closeButton, QtCore.SIGNAL(
            "clicked()"), PoolEditor.close)
        QtCore.QMetaObject.connectSlotsByName(PoolEditor)

    def retranslateUi(self, PoolEditor):
        PoolEditor.setWindowTitle(QtGui.QApplication.translate(
            "PoolEditor", "Create Pool", None, QtGui.QApplication.UnicodeUTF8))
        self.instanceNameLabel.setText(QtGui.QApplication.translate(
            "PoolEditor", "Instance name", None, QtGui.QApplication.UnicodeUTF8))
        self.poolDeviceNameLabel.setText(QtGui.QApplication.translate(
            "PoolEditor", "Pool device name", None, QtGui.QApplication.UnicodeUTF8))
        self.aliasLabel.setText(QtGui.QApplication.translate(
            "PoolEditor", "Alias (optional)", None, QtGui.QApplication.UnicodeUTF8))
        self.poolVersionLabel.setText(QtGui.QApplication.translate(
            "PoolEditor", "Pool version", None, QtGui.QApplication.UnicodeUTF8))
        self.poolDeviceNameCheckBox.setText(QtGui.QApplication.translate(
            "PoolEditor", "Automatic", None, QtGui.QApplication.UnicodeUTF8))
        self.aliasCheckBox.setText(QtGui.QApplication.translate(
            "PoolEditor", "Automatic", None, QtGui.QApplication.UnicodeUTF8))
        self.poolVersionCheckBox.setText(QtGui.QApplication.translate(
            "PoolEditor", "Automatic", None, QtGui.QApplication.UnicodeUTF8))
        self.groupBox.setTitle(QtGui.QApplication.translate(
            "PoolEditor", "PoolPath", None, QtGui.QApplication.UnicodeUTF8))
        self.createButton.setText(QtGui.QApplication.translate(
            "PoolEditor", "Create", None, QtGui.QApplication.UnicodeUTF8))
        self.closeButton.setText(QtGui.QApplication.translate(
            "PoolEditor", "Close", None, QtGui.QApplication.UnicodeUTF8))


if __name__ == "__main__":
    import sys
    app = QtGui.QApplication(sys.argv)
    PoolEditor = QtGui.QDialog()
    ui = Ui_PoolEditor()
    ui.setupUi(PoolEditor)
    PoolEditor.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from unittest import TestCase

from sardana.pool.test.base import BasePoolTestCase


class PseudoCounterTestCase(BasePoolTestCase, TestCase):
    """TestCase with PseudoCounter integration tests."""

    def setUp(self):
        """Create IoverI0 pseudo counter based on two dummy counters"""
        BasePoolTestCase.setUp(self)
        ctctrl = self.createController("ctctrl1",
                                       "DummyCounterTimerController",
                                       "DummyCounterTimerController")
        ct1 = self.ct1 = self.createCTElement(ctctrl, "ct1", 1)
        ct2 = self.ct2 = self.createCTElement(ctctrl, "ct2", 2)
        pcctrl = self.createController("pcctrl1", "IoverI0", "IoverI0")
        self.pc = self.createPCElement(pcctrl, "pc1", 1, (ct1.id, ct2.id))

    def test_pseudocounter_calc_buffer(self):
        """Simulate acquisition by filling the value buffer of the counters
        and test that the last_value_chunk of the pseudo counter is correctly
        filled.
        """
        pc_value_buffer = self.pc.get_value_buffer()
        self.ct1.extend_value_buffer([1., 2., 3.])
        self.ct2.append_value_buffer(10.)
        self.assertEqual(len(pc_value_buffer.last_chunk), 1)
        self.assertEqual(pc_value_buffer.last_chunk[0].value, 0.1)
        self.ct1.extend_value_buffer([4., 5., 6.])
        self.ct2.append_value_buffer(10., idx=3)
        self.assertEqual(len(pc_value_buffer.last_chunk), 1)
        self.assertEqual(pc_value_buffer.last_chunk[3].value, 0.4)
        self.ct2.append_value_buffer(10., idx=5)
        self.assertEqual(len(pc_value_buffer.last_chunk), 1)
        self.assertEqual(pc_value_buffer.last_chunk[5].value, 0.6)
        self.ct1.extend_value_buffer([7., 8.])
        self.ct1.extend_value_buffer([9.])
        self.ct1.extend_value_buffer([10.])
        self.ct2.append_value_buffer(10., idx=8)
        self.assertEqual(len(pc_value_buffer.last_chunk), 1)
        self.assertEqual(pc_value_buffer.last_chunk[8].value, 0.9)
        self.ct2.append_value_buffer(10., idx=9)
        self.assertEqual(len(pc_value_buffer.last_chunk), 1)
        self.assertEqual(pc_value_buffer.last_chunk[9].value, 1)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

__all__ = ['AttributeListener']

import numpy
import threading


class AttributeListener(object):

    def __init__(self, dtype=object, attr_name="valuebuffer"):
        self.data = {}
        self.dtype = dtype
        self.attr_name = attr_name
        self.data_lock = threading.RLock()

    def event_received(self, *args, **kwargs):
        # s - type: sardana.sardanavalue.SardanaValue
        # t - type: sardana.sardanaevent.EventType
        # v - type: sardana.sardanaattribute.SardanaAttribute e.g.
        #           sardana.pool.poolbasechannel.Value
        s, t, v = args
        if t.name.lower() != self.attr_name:
            return
        # obtaining sardana element e.g. exp. channel (the attribute owner)
        obj_name = s.name
        # obtaining the SardanaValue(s) either from the value_chunk (in case
        # of buffered attributes) or from the value in case of normal
        # attributes
        chunk = v
        idx = list(chunk.keys())
        value = [sardana_value.value for sardana_value in list(chunk.values())]
        # filling the measurement records
        with self.data_lock:
            channel_data = self.data.get(obj_name, [])
            expected_idx = len(channel_data)
            pad = [None] * (idx[0] - expected_idx)
            channel_data.extend(pad + value)
            self.data[obj_name] = channel_data

    def get_table(self):
        '''Construct a table-like array with padded  channel data as columns.
        Return the '''
        with self.data_lock:
            max_len = max([len(d) for d in list(self.data.values())])
            dtype_spec = []
            table = []
            for k in sorted(self.data.keys()):
                v = self.data[k]
                v.extend([None] * (max_len - len(v)))
                table.append(v)
                dtype_spec.append((k, self.dtype))
            a = numpy.array(list(zip(*table)), dtype=dtype_spec)
            return a



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for experiment channels"""

__all__ = ["Value", "PoolBaseChannel"]

__docformat__ = 'restructuredtext'

from sardana.sardanadefs import AttrQuality
from sardana.sardanaattribute import SardanaAttribute
from sardana.sardanabuffer import SardanaBuffer
from sardana.pool.poolelement import PoolElement
from sardana.pool.poolacquisition import PoolAcquisitionSoftware,\
    get_timerable_items
from sardana.pool.poolmeasurementgroup import ChannelConfiguration,\
    ControllerConfiguration
from sardana.sardanaevent import EventType
from sardana.pool import AcqSynch, AcqMode

class ValueBuffer(SardanaBuffer):

    def is_value_required(self, idx):
        """Check whether any of pseudo elements still still requires
        this value.

        :param idx: value's index
        :type idx: int
        :return: whether value is required or can be freely removed
        :rtype: bool
        """
        for element in self.obj.get_pseudo_elements():
            if element.get_value_buffer().next_idx <= idx:
                return True
        return False


class Value(SardanaAttribute):

    def __init__(self, *args, **kwargs):
        super(Value, self).__init__(*args, **kwargs)

    def update(self, cache=True, propagate=1):
        if not cache or not self.has_value():
            value = self.obj.read_value()
            self.set_value(value, propagate=propagate)


class ValueRefBuffer(SardanaBuffer):
    """Buffer for value refs.

    .. note::
        The ValueRefBuffer class has been included in Sardana on a provisional
        basis. Backwards incompatible changes (up to and including removal
        of the class) may occur if deemed necessary by the core developers.
    """
    pass


class ValueRef(SardanaAttribute):
    """Value ref attribute.

    .. note::
        The ValueRef class has been included in Sardana on a provisional
        basis. Backwards incompatible changes (up to and including removal
        of the class) may occur if deemed necessary by the core developers.
    """
    pass


class PoolBaseChannel(PoolElement):
    """Base class for Pool experimental channel

    .. todo:: Value ref API should be exposed only on the channels which
        belongs to *referable* controllers.
    """

    ValueAttributeClass = Value
    ValueBufferClass = ValueBuffer
    ValueRefAttributeClass = ValueRef
    ValueRefBufferClass = ValueRefBuffer
    AcquisitionClass = PoolAcquisitionSoftware

    def __init__(self, **kwargs):
        PoolElement.__init__(self, **kwargs)
        self._value = self.ValueAttributeClass(self, listeners=self.on_change)
        self._value_buffer = self.ValueBufferClass(self,
                                                   listeners=self.on_change)
        self._value_ref = self.ValueRefAttributeClass(
            self, listeners=self.on_change)
        self._value_ref_buffer = self.ValueRefBufferClass(
            self, listeners=self.on_change)
        self._value_ref_pattern = None
        self._value_ref_enabled = None
        self._pseudo_elements = []
        if self.AcquisitionClass is not None:
            acq_name = "%s.Acquisition" % self._name
            self.set_action_cache(self.AcquisitionClass(self, name=acq_name))
        self._integration_time = 0

    def has_pseudo_elements(self):
        """Informs whether this channel forms part of any pseudo element
        e.g. pseudo counter.

        :return: has pseudo elements
        :rtype: bool
        """
        return len(self._pseudo_elements) > 0

    def get_pseudo_elements(self):
        """Returns list of pseudo elements e.g. pseudo counters that this
        channel belongs to.

        :return: pseudo elements
        :rtype: seq<:class:`~sardana.pool.poolpseudocounter.PoolPseudoCounter`>
        """
        return self._pseudo_elements

    def add_pseudo_element(self, element):
        """Adds pseudo element e.g. pseudo counter that this channel
        belongs to.

        :param element: pseudo element
        :type element:
            :class:`~sardana.pool.poolpseudocounter.PoolPseudoCounter`
        """
        if not self.has_pseudo_elements():
            self.get_value_buffer().persistent = True
        self._pseudo_elements.append(element)

    def remove_pseudo_element(self, element):
        """Removes pseudo element e.g. pseudo counters that this channel
        belongs to.

        :param element: pseudo element
        :type element:
            :class:`~sardana.pool.poolpseudocounter.PoolPseudoCounter`
        """

        self._pseudo_elements.remove(element)
        if not self.has_pseudo_elements():
            self.get_value_buffer().persistent = False

    def get_value_attribute(self):
        """Returns the value attribute object for this experiment channel

        :return: the value attribute
        :rtype: :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        return self._value

    def get_value_buffer(self):
        """Returns the value attribute object for this experiment channel

        :return: the value attribute
        :rtype: :class:`~sardana.sardanaattribute.SardanaBuffer`"""
        return self._value_buffer

    def get_value_ref_attribute(self):
        """Returns the value attribute object for this experiment channel

        .. note::
            The get_value_ref_attribute method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :return: the value attribute
        :rtype: :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        return self._value_ref

    def get_value_ref_buffer(self):
        """Returns the value attribute object for this experiment channel

        .. note::
            The get_value_ref_buffer method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :return: the value attribute
        :rtype: :class:`~sardana.sardanaattribute.SardanaBuffer`"""
        return self._value_ref_buffer

    # --------------------------------------------------------------------------
    # Event forwarding
    # --------------------------------------------------------------------------

    def on_change(self, evt_src, evt_type, evt_value):
        # forward all events coming from attributes to the listeners
        self.fire_event(evt_type, evt_value)

    # --------------------------------------------------------------------------
    # default acquisition channel
    # --------------------------------------------------------------------------

    def get_default_attribute(self):
        return self.get_value_attribute()

    # --------------------------------------------------------------------------
    # acquisition
    # --------------------------------------------------------------------------

    def get_acquisition(self):
        return self.get_action_cache()

    acquisition = property(get_acquisition, doc="acquisition object")

    # --------------------------------------------------------------------------
    # value
    # --------------------------------------------------------------------------

    def read_value(self):
        """Reads the channel value from hardware.

        :return:
            a :class:`~sardana.sardanavalue.SardanaValue` containing the channel
            value
        :rtype:
            :class:`~sardana.sardanavalue.SardanaValue`"""
        return self.acquisition.read_value()[self]

    def put_value(self, value, quality=AttrQuality.Valid, propagate=1):
        """Sets a value.

        :param value:
            the new value
        :type value:
            :class:`~sardana.sardanavalue.SardanaValue`
        :param quality:
            the new quality
        :type quality:
            :class:`~sardana.AttrQuality`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        """
        val_attr = self._value
        val_attr.set_quality(quality)
        val_attr.set_value(value, propagate=propagate)
        return val_attr

    def get_value(self, cache=True, propagate=1):
        """Returns the channel value.

        :param cache:
            if ``True`` (default) return value in cache, otherwise read value
            from hardware
        :type cache:
            bool
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        :return:
            the channel value
        :rtype:
            :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        return self._get_value(cache=cache, propagate=propagate)

    def _get_value(self, cache=True, propagate=1):
        value = self.get_value_attribute()
        value.update(cache=cache, propagate=propagate)
        return value

    def set_value(self, value):
        """Starts an acquisition on this channel

        :param value:
            the value to count
        :type value:
            :class:`~numbers.Number`"""
        return self._set_value(value)

    def _set_value(self, value):
        self.start_acquisition(value)

    value = property(get_value, set_value, doc="channel value")

    # ------------------------------------------------------------------------
    # value buffer
    # ------------------------------------------------------------------------

    def extend_value_buffer(self, values, idx=None, propagate=1):
        """Extend value buffer with new values assigning them consecutive
        indexes starting with idx. If idx is omitted, then the new values will
        be added right after the last value in the buffer. Also update the read
        value of the attribute with the last element of values.

        :param values:
            values to be added to the buffer
        :type values:
            :class:`~sardana.sardanavalue.SardanaValue`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int
        """
        if len(values) == 0:
            return
        # fill value buffer
        val_buffer = self._value_buffer
        val_buffer.extend(values, idx)
        # update value attribute
        val_attr = self._value
        val_attr.set_value(values[-1], propagate=propagate)
        return val_buffer

    def append_value_buffer(self, value, idx=None, propagate=1):
        """Extend value buffer with new values assigning them consecutive
        indexes starting with idx. If idx is omitted, then the new value will
        be added with right after the last value in the buffer. Also update
        the read value.

        :param value:
            value to be added to the buffer
        :type value:
            :class:`~sardana.sardanavalue.SardanaValue`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int
        """
        # fill value buffer
        val_buffer = self._value_buffer
        val_buffer.append(value, idx)
        # update value attribute
        val_attr = self._value
        val_attr.set_value(value, propagate=propagate)
        return val_buffer

    def clear_value_buffer(self):
        val_attr = self._value_buffer
        val_attr.clear()

    # ------------------------------------------------------------------------
    # value ref
    # ------------------------------------------------------------------------

    def read_value_ref(self):
        """Reads the channel value ref from hardware.

        .. note::
            The read_value_ref method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :return:
            a *sardana value* containing the channel value
        :rtype:
            :class:`~sardana.sardanavalue.SardanaValue`"""
        return self.acquisition.read_value_ref()[self]

    def put_value_ref(self, value_ref, propagate=1):
        """Sets a value ref.

        .. note::
            The put_value_ref method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :param value_ref:
            the new value
        :type value_ref:
            :class:`~sardana.sardanavalue.SardanaValue`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        """
        val_ref_attr = self._value_ref
        val_ref_attr.set_value(value_ref, propagate=propagate)
        return val_ref_attr

    def get_value_ref(self):
        """Returns the channel value ref.

        .. note::
            The get_value_ref method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :return:
            the channel value
        :rtype:
            :class:`~sardana.SardanaValue`"""
        value_ref_attr = self.get_value_ref_attribute()
        if not value_ref_attr.has_value():
            raise Exception("ValueRef not available: no successful"
                            " acquisition done so far!")
        return value_ref_attr.value_obj

    value_ref = property(get_value_ref, doc="channel value ref")

    # ------------------------------------------------------------------------
    # value ref buffer
    # ------------------------------------------------------------------------

    def extend_value_ref_buffer(self, value_refs, idx=None, propagate=1):
        """Extend value ref buffer with new values assigning them consecutive
        indexes starting with idx. If idx is omitted, then the new values will
        be added right after the last value ref in the buffer. Also update the
        read value ref of the attribute with the last element of values.

        .. note::
            The extend_value_ref_buffer method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :param value_refs:
            values to be added to the buffer
        :type value_refs:
            :class:`~sardana.sardanavalue.SardanaValue`
        :param idx:
            index at which to append the value_refs
        :type idx: :obj:`int`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int
        """
        if len(value_refs) == 0:
            return
        # fill value ref buffer
        val_ref_buffer = self._value_ref_buffer
        val_ref_buffer.extend(value_refs, idx)
        # update value ref attribute
        val_ref_attr = self._value_ref
        val_ref_attr.set_value(value_refs[-1], propagate=propagate)
        return val_ref_buffer

    def append_value_ref_buffer(self, value_ref, idx=None, propagate=1):
        """Append new value ref to the value ref buffer at the idx position.

        If idx is omitted, then the new value ref will be added right
        after the last value ref in the buffer.
        Also update the read value ref.

        .. note::
            The append_value_ref_buffer method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :param value_ref:
            value ref to be added to the buffer
        :type value_ref:
            :class:`~sardana.sardanavalue.SardanaValue`
        :param idx:
            index at which to append the value_ref
        :type idx: :obj:`int`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int
        """
        # fill value ref buffer
        val_ref_buffer = self._value_ref_buffer
        val_ref_buffer.append(value_ref, idx)
        # update value ref attribute
        val_ref_attr = self._value_ref
        val_ref_attr.set_value(value_ref, propagate=propagate)
        return val_ref_buffer

    def clear_value_ref_buffer(self):
        """Clear value ref buffer.

        .. note::
            The clear_value_ref_buffer method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.
        """
        val_ref_attr = self._value_ref_buffer
        val_ref_attr.clear()

    # ------------------------------------------------------------------------
    # value ref pattern
    # ------------------------------------------------------------------------

    def get_value_ref_pattern(self):
        """Returns the channel value reference pattern.

        .. note::
            The get_value_ref_pattern method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :return:
            the channel value
        :rtype:
            :obj:`str`
        """
        return self._value_ref_pattern

    def set_value_ref_pattern(self, value_ref_pattern, propagate=1):
        """Set a value reference pattern in the kernel (not in the hardware).

        .. note::
            The set_value_ref_pattern method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :param value_ref_pattern:
            the new value reference pattern
        :type value_ref_pattern:
            :obj:`str`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            :obj:`int`
        """
        self._set_value_ref_pattern(value_ref_pattern, propagate=propagate)

    def _set_value_ref_pattern(self, value_ref_pattern, propagate=1):
        self._value_ref_pattern = value_ref_pattern
        if not propagate:
            return
        self.fire_event(
            EventType("value_ref_pattern", priority=propagate),
            value_ref_pattern)

    value_ref_pattern = property(get_value_ref_pattern,
                                 set_value_ref_pattern,
                                 doc="channel value reference pattern")

    # ------------------------------------------------------------------------
    # value ref enabled
    # ------------------------------------------------------------------------

    def is_value_ref_enabled(self):
        """Check if value reference is enabled.

        .. note::
            The is_value_ref_enabled method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :return:
            the channel value
        :rtype:
            :obj:`bool`
        """
        return self._value_ref_enabled

    def set_value_ref_enabled(self, value_ref_enabled, propagate=1):
        """Set value reference enabled flag in the kernel (not in the
        hardware).

        .. note::
            The set_value_ref_enabled method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.

        :param value_ref_enabled:
            the new value reference enabled
        :type value_ref_enabled:
            :obj:`bool`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            :obj:`int`
        """
        self._set_value_ref_enabled(value_ref_enabled, propagate=propagate)

    def _set_value_ref_enabled(self, value_ref_enabled, propagate=1):
        self._value_ref_enabled = value_ref_enabled
        if not propagate:
            return
        self.fire_event(
            EventType("value_ref_enabled", priority=propagate),
            value_ref_enabled)

    value_ref_enabled = property(is_value_ref_enabled,
                                 set_value_ref_enabled,
                                 doc="channel value reference enabled")

    def is_referable(self):
        """Check if channel has referable capability.

        .. note::
            The is_referable method has been included in Sardana on
            a provisional basis. Backwards incompatible changes (up to and
            including removal of the class) may occur if deemed necessary by
            the core developers.
        """
        return self.controller.is_referable()

    # --------------------------------------------------------------------------
    # integration time
    # --------------------------------------------------------------------------

    def get_integration_time(self):
        """Return the integration time for this object.

        :return: the current integration time
        :rtype: :obj:`float`"""
        return self._integration_time

    def set_integration_time(self, integration_time, propagate=1):
        """Set the integration time for this object.

        :param integration_time: integration time in seconds to set
        :type integration_time: :obj:`float`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: :obj:`int`
        """
        if integration_time == self._integration_time:
            # integration time is not changed. Do nothing
            return
        self._integration_time = integration_time
        if not propagate:
            return
        self.fire_event(EventType("integration_time", priority=propagate),
                        integration_time)

    integration_time = property(get_integration_time, set_integration_time,
                                doc="channel integration time")

    def _prepare(self):
        # TODO: think of implementing the preparation in the software
        # acquisition action, similarly as it is done for the global
        # acquisition action
        if self.is_referable():
            self.controller.set_axis_par(self.axis, "value_ref_enabled",
                                         self.value_ref_enabled)
            if self.value_ref_enabled:
                self.controller.set_axis_par(self.axis, "value_ref_pattern",
                                             self.value_ref_pattern)

    def start_acquisition(self):
        msg = "{0} does not support independent acquisition".format(
            self.__class__.__name__)
        raise NotImplementedError(msg)


class PoolTimerableChannel(PoolBaseChannel):

    def __init__(self, **kwargs):
        PoolBaseChannel.__init__(self, **kwargs)
        # TODO: part of the configuration could be moved to the base class
        # so other experimental channels could reuse it
        self._conf_channel = ChannelConfiguration(self)
        self._conf_ctrl = ControllerConfiguration(self.controller)
        self._conf_ctrl.add_channel(self._conf_channel)
        self._timer = "__default"
        self._conf_timer = None

    # ------------------------------------------------------------------------
    # timer
    # ------------------------------------------------------------------------

    def get_timer(self):
        """Return the timer for this object.

        :return: the current timer
        :rtype: :obj:`str`
        """
        return self._timer

    def set_timer(self, timer, propagate=1):
        """Set timer for this object.

        :param timer: new timer to set
        :type timer: :obj:`str`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: :obj:`int`
        """
        if timer == self._timer:
            return
        if self._conf_timer is not None:
            timer_elem = self._conf_timer.element
            if timer_elem is not self:
                self.acquisition.remove_element(timer_elem)
                self._conf_ctrl.remove_channel(self._conf_timer)
        self._timer = timer
        self._conf_timer = None
        if not propagate:
            return
        self.fire_event(EventType("timer", priority=propagate), timer)

    timer = property(get_timer, set_timer,
                     doc="timer for the timerable channel")

    def _configure_timer(self):
        timer = self.timer
        if timer == "__self":
            conf_timer = self._conf_channel
        else:
            ctrl = self.get_controller()
            if timer == "__default":
                axis = ctrl.get_default_timer()
                if axis is None:
                    msg = "default_timer not defined in controller"
                    raise ValueError(msg)
                timer_elem = ctrl.get_element(axis=axis)
            else:
                timer_elem = self.pool.get_element_by_name(timer)
            if timer_elem is self:
                conf_timer = self._conf_channel
            else:
                self.acquisition.add_element(timer_elem)
                conf_timer = ChannelConfiguration(timer_elem)
                self._conf_ctrl.add_channel(conf_timer)
        self._conf_ctrl.timer = conf_timer
        self._conf_timer = conf_timer

    def start_acquisition(self):
        """Start software triggered acquisition"""
        self._aborted = False
        self._stopped = False
        if self._simulation_mode:
            return
        if self.timer is None:
            msg = "no timer configured - acquisition is not possible"
            raise RuntimeError(msg)
        self._prepare()
        ctrls, master = get_timerable_items(
            [self._conf_ctrl], self._conf_timer, AcqMode.Timer)
        self.acquisition.run(ctrls, self.integration_time, master, None)

    def _prepare(self):
        # TODO: think of implementing the preparation in the software
        # acquisition action, similarly as it is done for the global
        # acquisition action
        PoolBaseChannel._prepare(self)
        if self._conf_timer is None:
            self._configure_timer()
        axis = self._conf_timer.axis
        repetitions = 1
        latency = 0
        nb_starts = 1
        self.controller.set_ctrl_par("synchronization",
                                     AcqSynch.SoftwareTrigger)
        self.controller.ctrl.PrepareOne(axis, self.integration_time,
                                        repetitions, latency, nb_starts)



#!/usr/bin/env python

#
#
# This file is part of Sardana
#
# http://www.sardana-controls.org/
#
# Copyright 2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
#
#


"""
This module provides a recorder for NXscans implemented with h5py (no nxs)
"""

__all__ = ["NXscanH5_FileRecorder"]

__docformat__ = 'restructuredtext'

import os
import re
import posixpath
from datetime import datetime
import numpy
import h5py

from sardana.sardanautils import is_pure_str
from sardana.taurus.core.tango.sardana import PlotType
from sardana.macroserver.scan.recorder import BaseFileRecorder, SaveModes

VDS_available = True
try:
    h5py.VirtualSource
except AttributeError:
    VDS_available = False


def timedelta_total_seconds(timedelta):
    """Equivalent to timedelta.total_seconds introduced with python 2.7."""
    return (
        timedelta.microseconds + 0.0
        + (timedelta.seconds + timedelta.days * 24 * 3600)
        * 10 ** 6) / 10 ** 6


class NXscanH5_FileRecorder(BaseFileRecorder):

    """
    Saves data to a nexus file that follows the NXscan application definition
    (This is a pure h5py implementation that does not depend on the nxs module)
    """
    formats = {'h5': '.h5'}
    # from http://docs.h5py.org/en/latest/strings.html
    str_dt = h5py.special_dtype(vlen=str)  # Variable-length UTF-8
    byte_dt = h5py.special_dtype(vlen=bytes)  # Variable-length UTF-8
    supported_dtypes = ('float32', 'float64', 'int8',
                        'int16', 'int32', 'int64', 'uint8',
                        'uint16', 'uint32',
                        'uint64', str_dt, byte_dt)
    _dataCompressionRank = -1

    def __init__(self, filename=None, macro=None, overwrite=False, **pars):
        BaseFileRecorder.__init__(self, **pars)

        self.macro = macro
        self.overwrite = overwrite
        if filename:
            self.setFileName(filename)

        self.currentlist = None
        self._nxclass_map = {}
        self.entryname = 'entry'

        scheme = r'([A-Za-z][A-Za-z0-9\.\+\-]*)'
        authority = (r'//(?P<host>([\w\-_]+\.)*[\w\-_]+)'
                     + r'(:(?P<port>\d{1,5}))?')
        path = (r'((?P<filepath>(/(//+)?([A-Za-z]:/)?([\w.\-_]+/)*'
                + r'[\w.\-_]+.(h5|hdf5|\w+))))'
                + r'(::(?P<dataset>(([\w.\-_]+/)*[\w.\-_]+)))?')
        pattern = ('^(?P<scheme>%(scheme)s):'
                   + '((?P<authority>%(authority)s)'
                   + '($|(?=[/#?])))?(?P<path>%(path)s)$')
        self.pattern = pattern % dict(scheme=scheme, authority=authority,
                                      path=path)

    def getFormat(self):
        return 'HDF5::NXscan'

    def setFileName(self, filename):
        if self.fd is not None:
            self.fd.close()
        self.filename = filename
        self.currentlist = None

    def sanitizeName(self, name):
        """It returns a version of the given name that can be used as a python
        variable (and conforms to NeXus best-practices for dataset names)
        """
        # make sure the name does not start with a digit
        if name[0].isdigit():
            name = "_%s" % name
        # substitute whitespaces by underscores and remove other
        # non-alphanumeric characters
        return "".join(
            x for x in name.replace(' ', '_') if x.isalnum() or x == '_')

    def _openFile(self, fname):
        """Open the file with given filename (create if it does not exist)
        Populate the root of the file with some metadata from the NXroot
        definition"""
        if os.path.exists(fname):
            fd = h5py.File(fname, mode='r+')
        else:
            fd = h5py.File(fname, mode='w-')
            fd.attrs['NX_class'] = 'NXroot'
            fd.attrs['file_name'] = fname
            fd.attrs['file_time'] = datetime.now().isoformat()
            fd.attrs['creator'] = self.__class__.__name__
            fd.attrs['HDF5_Version'] = h5py.version.hdf5_version
            fd.attrs['h5py_version'] = h5py.version.version
        return fd

    def _startRecordList(self, recordlist):

        if self.filename is None:
            return

        self.currentlist = recordlist
        env = self.currentlist.getEnviron()
        serialno = env['serialno']
        self._dataCompressionRank = env.get('DataCompressionRank',
                                            self._dataCompressionRank)

        # open/create the file and store its descriptor
        self.fd = self._openFile(self.filename)

        # create an entry for this scan using the scan serial number
        self.entryname = 'entry%d' % serialno
        try:
            nxentry = self.fd.create_group(self.entryname)
        except ValueError:
            # Warn and abort
            if self.entryname in list(self.fd.keys()):
                msg = ('{ename:s} already exists in {fname:s}. '
                       'Aborting macro to prevent data corruption.\n'
                       'This is likely caused by a wrong ScanID\n'
                       'Possible workarounds:\n'
                       '  * first, try re-running this macro (the ScanID '
                       'may be automatically corrected)\n'
                       '  * if not, try changing ScanID with senv, or...\n'
                       '  * change the file name ({ename:s} will be in both '
                       'files containing different data)\n'
                       '\nPlease report this problem.'
                       ).format(ename=self.entryname, fname=self.filename)
                raise RuntimeError(msg)
            else:
                raise
        nxentry.attrs['NX_class'] = 'NXentry'

        # adapt the datadesc to the NeXus requirements
        self.datadesc = []
        for dd in env['datadesc']:
            dd = dd.clone()
            dd.label = self.sanitizeName(dd.label)
            if not hasattr(dd, "value_ref_enabled"):
                dd.value_ref_enabled = False
            if dd.dtype == 'bool':
                dd.dtype = 'int8'
                self.debug('%r will be stored with type=%r', dd.name, dd.dtype)
            elif dd.dtype == 'str':
                dd.dtype = NXscanH5_FileRecorder.str_dt
            if dd.value_ref_enabled:
                # substitute original data (image or spectrum) type and shape
                # since we will receive references instead
                dd.dtype = NXscanH5_FileRecorder.str_dt
                dd.shape = tuple()
                self.debug('%r will be stored with type=%r', dd.name, dd.dtype)
            if dd.dtype in self.supported_dtypes:
                self.datadesc.append(dd)
            else:
                msg = '%r will not be stored. Reason: %r not supported'
                self.warning(msg, dd.name, dd.dtype)

        # make a dictionary out of env['instrumentlist']
        # (use fullnames -paths- as keys)
        self._nxclass_map = {}
        for inst in env.get('instrumentlist', []):
            self._nxclass_map[nxentry.name + inst.getFullName()] = inst.klass
        if self._nxclass_map is {}:
            self.warning('Missing information on NEXUS structure. '
                         + 'Nexus Tree will not be created')

        self.debug('Starting new recording %d on file %s', serialno,
                   self.filename)

        # populate the entry with some data
        nxentry.create_dataset('definition', data='NXscan')
        import sardana.release
        program_name = '%s (%s)' % (sardana.release.name,
                                    self.__class__.__name__)
        _pname = nxentry.create_dataset('program_name', data=program_name)
        _pname.attrs['version'] = sardana.release.version
        nxentry.create_dataset('start_time', data=env['starttime'].isoformat())
        timedelta = (env['starttime'] - datetime(1970, 1, 1))
        try:
            _epoch = timedelta.total_seconds()
        except AttributeError:
            # for python 2.6 compatibility
            _epoch = timedelta_total_seconds(timedelta)
        nxentry.attrs['epoch'] = _epoch
        nxentry.create_dataset('title', data=env['title'])
        nxentry.create_dataset('entry_identifier', data=str(env['serialno']))

        _usergrp = nxentry.create_group('user')
        _usergrp.attrs['NX_class'] = 'NXuser'
        _usergrp.create_dataset('name', data=env['user'])

        # prepare the 'measurement' group
        _meas = nxentry.create_group('measurement')
        _meas.attrs['NX_class'] = 'NXcollection'
        if self.savemode == SaveModes.Record:
            # create extensible datasets
            for dd in self.datadesc:
                shape = ([0] + list(dd.shape))
                _ds = _meas.create_dataset(
                    dd.label,
                    dtype=dd.dtype,
                    shape=shape,
                    maxshape=([None] + list(dd.shape)),
                    chunks=(1,) + tuple(dd.shape),
                    compression=self._compression(shape)
                )
                if hasattr(dd, 'data_units'):
                    _ds.attrs['units'] = dd.data_units

        else:
            # leave the creation of the datasets to _writeRecordList
            # (when we actually know the length of the data to write)
            pass

        self._createPreScanSnapshot(env)

        self.fd.flush()

    def _compression(self, shape, compfilter='gzip'):
        """
        Returns `compfilter` (the name of the compression filter) to use
        (or None if no compression is recommended), based on the given shape
        and the self._dataCompressionRank thresshold.
        By default, `compfilter` is set to `'gzip'`
        """
        min_rank = self._dataCompressionRank
        if shape is None or min_rank < 0 or len(shape) < min_rank:
            compfilter = None
        elif len(shape) == 0:
            msg = "%s compression is not supported for scalar datasets"\
                  " - these datasets will not be compressed" % compfilter
            self.warning(msg)
            compfilter = None
        return compfilter

    def _createPreScanSnapshot(self, env):
        """
        Write the pre-scan snapshot in "<entry>/measurement/pre_scan_snapshot".
        Also link to the snapshot datasets from the <entry>/measurement group
        """
        _meas = self.fd[posixpath.join(self.entryname, 'measurement')]
        self.preScanSnapShot = env.get('preScanSnapShot', [])
        _snap = _meas.create_group('pre_scan_snapshot')
        _snap.attrs['NX_class'] = 'NXcollection'

        meas_keys = list(_meas.keys())

        for dd in self.preScanSnapShot:
            label = self.sanitizeName(dd.label)
            dtype = dd.dtype
            pre_scan_value = dd.pre_scan_value
            if dd.dtype == 'bool':
                dtype = 'int8'
                pre_scan_value = numpy.int8(dd.pre_scan_value)
                self.debug('Pre-scan snapshot of %s will be stored as type %s',
                           dd.name, dtype)
            elif dd.dtype == 'str':
                dd.dtype = NXscanH5_FileRecorder.str_dt

            if dtype in self.supported_dtypes:
                _ds = _snap.create_dataset(
                    label,
                    data=pre_scan_value,
                    compression=self._compression(dd.shape)
                )
                # link to this dataset also from the measurement group
                if label not in meas_keys:
                    _meas[label] = _ds
            else:
                self.warning(('Pre-scan snapshot of %s will not be stored. '
                              + 'Reason: type %s not supported'),
                             dd.name, dtype)

    def _writeRecord(self, record):
        if self.filename is None:
            return
        _meas = self.fd[posixpath.join(self.entryname, 'measurement')]

        for dd in self.datadesc:
            if dd.name in record.data:
                data = record.data[dd.name]
                _ds = _meas[dd.label]
                if data is None:
                    data = numpy.zeros(dd.shape, dtype=dd.dtype)
                # skip NaN if value reference is enabled
                if dd.value_ref_enabled and not is_pure_str(data):
                    continue
                elif not hasattr(data, 'shape'):
                    data = numpy.array([data], dtype=dd.dtype)
                elif dd.dtype != data.dtype.name:
                    self.debug('%s casted to %s (was %s)',
                               dd.label, dd.dtype, data.dtype.name)
                    data = data.astype(dd.dtype)

                # resize the dataset and add the latest chunk
                if _ds.shape[0] <= record.recordno:
                    _ds.resize(record.recordno + 1, axis=0)

                # write the slab of data
                _ds[record.recordno, ...] = data

            else:
                self.debug('missing data for label %r', dd.label)
        self.fd.flush()

    def _endRecordList(self, recordlist):

        if self.filename is None:
            return

        self._populateInstrumentInfo()
        self._createNXData()

        env = self.currentlist.getEnviron()
        nxentry = self.fd[self.entryname]

        for dd, dd_env in zip(self.datadesc, env["datadesc"]):
            label = dd.label

            # If h5file scheme is used: Creation of a Virtual Dataset
            if dd.value_ref_enabled:
                measurement = nxentry['measurement']
                first_reference = measurement[label][0]

                group = re.match(self.pattern, first_reference)
                if group is None:
                    msg = 'Unsupported reference %s' % first_reference
                    self.warning(msg)
                    continue

                uri_groups = group.groupdict()
                if uri_groups['scheme'] != "h5file":
                    continue
                if not VDS_available:
                    msg = ("VDS not available in this version of h5py, "
                           "{0} will be stored as string reference")
                    msg.format(label)
                    self.warning(msg)
                    continue

                bk_label = "_" + label
                measurement[bk_label] = measurement[label]
                nb_points = measurement[label].size

                (dim_1, dim_2) = dd_env.shape
                layout = h5py.VirtualLayout(shape=(nb_points, dim_1, dim_2),
                                            dtype=dd_env.dtype)

                for i in range(nb_points):
                    reference = measurement[label][i]
                    group = re.match(self.pattern, reference)
                    if group is None:
                        msg = 'Unsupported reference %s' % first_reference
                        self.warning(msg)
                        continue
                    uri_groups = group.groupdict()
                    filename = uri_groups["filepath"]
                    dataset = uri_groups.get("dataset", "dataset")
                    vsource = h5py.VirtualSource(filename, dataset,
                                                 shape=(dim_1, dim_2))
                    layout[i] = vsource

                # Substitute dataset by Virtual Dataset in output file
                try:
                    del measurement[label]
                    measurement.create_virtual_dataset(
                        label, layout, fillvalue=-numpy.inf)
                except Exception as e:
                    msg = 'Could not create a Virtual Dataset. Reason: %r'
                    self.warning(msg, e)
                else:
                    del measurement[bk_label]

        nxentry.create_dataset('end_time', data=env['endtime'].isoformat())
        self.fd.flush()
        self.debug('Finishing recording %d on file %s:',
                   env['serialno'], self.filename)
        self.fd.close()
        self.currentlist = None

    def writeRecordList(self, recordlist):
        """Called when in BLOCK writing mode"""
        self._startRecordList(recordlist)
        _meas = self.fd[posixpath.join(self.entryname, 'measurement')]
        for dd in self.datadesc:
            shape = ([len(recordlist.records)] + list(dd.shape))
            _ds = _meas.create_dataset(
                dd.label,
                dtype=dd.dtype,
                shape=shape,
                chunks=(1,) + tuple(dd.shape),
                compression=self._compression(shape)
            )
            if hasattr(dd, 'data_units'):
                _ds.attrs['units'] = dd.data_units

            for record in recordlist.records:
                if dd.label in record.data:
                    _ds[record.recordno, ...] = record.data[dd.label]
                else:
                    self.debug('missing data for label %r in record %i',
                               dd.label, record.recordno)
        self._endRecordList(recordlist)

    def _populateInstrumentInfo(self):
        nxentry = self.fd[self.entryname]
        _meas = nxentry['measurement']
        _snap = _meas['pre_scan_snapshot']
        # create a link for each
        for dd in self.datadesc:
            # we only link if the instrument info is available
            if getattr(dd, 'instrument', None):
                try:
                    _ds = _meas[dd.label]
                    _instr = self._createNXpath(dd.instrument,
                                                prefix=nxentry.name)
                    _instr[os.path.basename(_ds.name)] = _ds
                except Exception as e:
                    msg = 'Could not create link to %r in %r. Reason: %r'
                    self.warning(msg, dd.label, dd.instrument, e)

        for dd in self.preScanSnapShot:
            if getattr(dd, 'instrument', None):
                label = self.sanitizeName(dd.label)
                try:
                    _ds = _snap[label]
                    _instr = self._createNXpath(dd.instrument,
                                                prefix=nxentry.name)
                    _instr[os.path.basename(_ds.name)] = _ds
                except Exception as e:
                    msg = 'Could not create link to %r in %r. Reason: %r'
                    self.warning(msg, label, dd.instrument, e)

    def _createNXData(self):
        """
        Creates groups of type NXdata by making links to the corresponding
        datasets
        """
        # classify by type of plot:
        plots1d = {}
        plots1d_names = {}
        i = 1
        for dd in self.datadesc:
            ptype = getattr(dd, 'plot_type', PlotType.No)
            if ptype == PlotType.No:
                continue
            elif ptype == PlotType.Spectrum:
                # converting the list into a colon-separated string
                axes = ':'.join(dd.plot_axes)
                if axes in plots1d:
                    plots1d[axes].append(dd)
                else:
                    plots1d[axes] = [dd]
                    # Note that datatesc ordering determines group name
                    # indexing
                    plots1d_names[axes] = 'plot_%i' % i
                    i += 1
            else:
                continue  # TODO: implement support for images and other

        nxentry = self.fd[self.entryname]
        _meas = nxentry['measurement']

        # write the 1D NXdata group
        for axes, v in list(plots1d.items()):
            _nxdata = nxentry.create_group(plots1d_names[axes])
            _nxdata.attrs['NX_class'] = 'NXdata'

            # write the signals
            for i, dd in enumerate(v):
                # link the datasets
                _ds = _nxdata[dd.label] = _meas[dd.label]
                # add attrs
                _ds.attrs['signal'] = min(i + 1, 2)
                _ds.attrs['axes'] = axes
                _ds.attrs['interpretation'] = 'spectrum'
            # write the axes
            for axis in axes.split(':'):
                try:
                    _nxdata[axis] = _meas[axis]
                except Exception:
                    self.warning('cannot create link for "%s". Skipping', axis)

    def _createNXpath(self, path, prefix=None):
        """
        Creates a path in the nexus file composed by nested data_groups
        with their corresponding NXclass attributes.

        This method creates the groups if they do not exist. If the
        path is given using `name:nxclass` notation, the given nxclass is
        used.
        Otherwise, the class name is obtained from self._nxclass_map values
        (and if not found, it defaults to NXcollection).

        It returns the tip of the branch (the last group created)
        """

        if prefix is None:
            # if prefix is None, use current entry if path is relative
            path = posixpath.join("/%s:NXentry" % self.entryname, path)
        else:
            # if prefix is given explicitly, assume that path is relative to it
            # even if path is absolute
            path = posixpath.join(prefix, path.lstrip('/'))

        grp = self.fd['/']
        for name in path[1:].split('/'):
            if ':' in name:
                name, nxclass = name.split(':')
            else:
                nxclass = None
            # open group (create if it does not exist)
            grp = grp.require_group(name)
            if 'NX_class' not in grp.attrs:
                if nxclass is None:
                    nxclass = self._nxclass_map.get(grp.name, 'NXcollection')
                grp.attrs['NX_class'] = nxclass
        return grp

    def _addCustomData(self, value, name, nxpath=None, dtype=None, **kwargs):
        """
        Apart from value and name, this recorder can use the following optional
        parameters:

        :param nxpath: (str) a nexus path (optionally using name:nxclass
                       notation for the group names). See the rules for
                       automatic nxclass resolution used by
                       :meth:`._createNXpath`. If None given, it defaults to
                       nxpath='custom_data:NXcollection'

        :param dtype: name of data type (it is inferred from value if not
                      given)
        """
        if nxpath is None:
            nxpath = 'custom_data:NXcollection'
        if dtype is None:
            if numpy.isscalar(value):
                dtype = numpy.dtype(type(value)).name
                if numpy.issubdtype(dtype, str):
                    dtype = 'char'
                if dtype == 'bool':
                    value, dtype = int(value), 'int8'
            else:
                value = numpy.array(value)
                dtype = value.dtype.name

        if dtype not in self.supported_dtypes and dtype != 'char':
            self.warning(
                'cannot write %r. Reason: unsupported data type', name)
            return

        # open the file if necessary
        fileClosed = self.fd is None or not hasattr(self.fd, 'mode')
        if fileClosed:
            self.fd = self._openFile(self.filename)

        # create the custom data group if it does not exist
        grp = self._createNXpath(nxpath)
        try:
            grp.create_dataset(name, data=value)
        except ValueError as e:
            msg = 'Error writing %s. Reason: %s' % (name, e)
            self.warning(msg)
            self.macro.warning(msg)

        # flush
        self.fd.flush()

        # leave the file as it was
        if fileClosed:
            self.fd.close()
